#include "../AudioEngine.h"
#include <SLES/OpenSLES.h>
#include <SLES/OpenSLES_Android.h>
#include <SLES/OpenSLES_AndroidConfiguration.h>
#include "../Utilities/Log.h"
#include "OVR_Voip_LowLevel.h"

namespace {
SLObjectItf engineObject;
SLEngineItf engineEngine;
SLObjectItf outputMixObject;

SLObjectItf playObj;
SLPlayItf playPlay;
SLAndroidSimpleBufferQueueItf playBufferQueue;
SLVolumeItf playVolume;

short PlayBuf[PlayBufferSize];
short MicBuf[MicBufferSize];

ovrMicrophoneHandle MicHandle = 0;

PlayCallback_t PlayCallback;
MicCallback_t MicCallback;

// Platform mic is annoying, doesnt always return 20ms for some reason.
uint32_t MicBufferUsed = 0;

void CheckSl(const char* title, const SLresult code) {
  if (code != SL_RESULT_SUCCESS) {
    OVR_FAIL("OpenSL error on %s: %i", title, code);
  }
}

void OpenSlEsPlayCallback(SLAndroidSimpleBufferQueueItf bq, void*) {
  if (PlayCallback) {
    PlayCallback(PlayBuf);
  } else {
    memset(PlayBuf, 0, sizeof(PlayBuf));
  }
  if (MicCallback) {
    size_t recvd = 0;
    do {
      recvd =
          ovr_Microphone_GetPCM(MicHandle, MicBuf + MicBufferUsed, (MicBufferSize - MicBufferUsed));
      MicBufferUsed += recvd;
      if (MicBufferUsed == MicBufferSize) {
        MicCallback(MicBuf);
        MicBufferUsed = 0;
      }
    } while (recvd != 0);
  }
  (*bq)->Enqueue(bq, PlayBuf, sizeof(PlayBuf));
}
} // namespace

void AudioEngineInitialize(PlayCallback_t playCallback, MicCallback_t micCallback) {
  PlayCallback = playCallback;
  MicCallback = micCallback;

  CheckSl("slCreateEngine", slCreateEngine(&engineObject, 0, NULL, 0, NULL, NULL));
  CheckSl("Realize", (*engineObject)->Realize(engineObject, SL_BOOLEAN_FALSE));
  CheckSl(
      "GetInterface", (*engineObject)->GetInterface(engineObject, SL_IID_ENGINE, &engineEngine));

  const SLInterfaceID ids2[1] = {};
  const SLboolean req2[1] = {};
  CheckSl(
      "CreateOutputMix",
      (*engineEngine)->CreateOutputMix(engineEngine, &outputMixObject, 0, ids2, req2));
  CheckSl("Realize", (*outputMixObject)->Realize(outputMixObject, SL_BOOLEAN_FALSE));

  // configure audio source
  unsigned channels = 2;
  unsigned speakers = SL_SPEAKER_FRONT_LEFT | SL_SPEAKER_FRONT_RIGHT;
  SLDataLocator_AndroidSimpleBufferQueue loc_bufq = {SL_DATALOCATOR_ANDROIDSIMPLEBUFFERQUEUE, 2};
  SLDataFormat_PCM format_pcm = {SL_DATAFORMAT_PCM,
                                 channels,
                                 SampleRate * 1000 /* crazy OpenSL rate scale */,
                                 SL_PCMSAMPLEFORMAT_FIXED_16,
                                 SL_PCMSAMPLEFORMAT_FIXED_16,
                                 speakers,
                                 SL_BYTEORDER_LITTLEENDIAN};
  SLDataSource audioSrc = {&loc_bufq, &format_pcm};

  // configure audio sink
  SLDataLocator_OutputMix loc_outmix = {SL_DATALOCATOR_OUTPUTMIX, outputMixObject};
  SLDataSink audioSnk = {&loc_outmix, NULL};

  // create audio player
  const SLInterfaceID ids[3] = {SL_IID_ANDROIDSIMPLEBUFFERQUEUE, SL_IID_VOLUME};
  const SLboolean req[3] = {SL_BOOLEAN_TRUE, SL_BOOLEAN_TRUE};
  CheckSl(
      "CreateAudioPlayer",
      (*engineEngine)
          ->CreateAudioPlayer(engineEngine, &playObj, &audioSrc, &audioSnk, 2, ids, req));
  CheckSl("Realize", (*playObj)->Realize(playObj, SL_BOOLEAN_FALSE));
  CheckSl("GetInterface", (*playObj)->GetInterface(playObj, SL_IID_PLAY, &playPlay));
  CheckSl(
      "GetInterface",
      (*playObj)->GetInterface(playObj, SL_IID_ANDROIDSIMPLEBUFFERQUEUE, &playBufferQueue));
  CheckSl(
      "RegisterCallback",
      (*playBufferQueue)->RegisterCallback(playBufferQueue, OpenSlEsPlayCallback, nullptr));
  CheckSl("GetInterface", (*playObj)->GetInterface(playObj, SL_IID_VOLUME, &playVolume));
  CheckSl("SetPlayState", (*playPlay)->SetPlayState(playPlay, SL_PLAYSTATE_PLAYING));

  memset(PlayBuf, 0, sizeof(PlayBuf));

  CheckSl("Enqueue", (*playBufferQueue)->Enqueue(playBufferQueue, PlayBuf, sizeof(PlayBuf)));

  MicBufferUsed = 0;
  MicHandle = ovr_Microphone_Create();
  ovr_Microphone_Start(MicHandle);
}

void AudioEngineShutdown() {
  ovr_Microphone_Stop(MicHandle);
  ovr_Microphone_Destroy(MicHandle);

  CheckSl("SetPlayState", (*playPlay)->SetPlayState(playPlay, SL_PLAYSTATE_STOPPED));
  playPlay = nullptr;

  CheckSl("ClearPlay", (*playBufferQueue)->Clear(playBufferQueue));
  playBufferQueue = nullptr;

  (*playObj)->Destroy(playObj);
  playObj = nullptr;
  (*outputMixObject)->Destroy(outputMixObject);
  outputMixObject = nullptr;
  (*engineObject)->Destroy(engineObject);
  engineObject = nullptr;
}
