#include "Application.h"
#include "AudioEngine.h"
#include "Shaders.h"
#include "Utilities/Log.h"

#include "OVR_Platform.h"

#include <cassert>
#include <cinttypes>

using namespace ovrUtilities;
using namespace std;

// These are Oculus Application Id's.
// To run these apps, you will need application Id's, fully setup,
// from your own organization. See this link for more information.
// https://developer.oculus.com/documentation/platform/latest/concepts/book-pgsg/
#ifdef ANDROID
static const char* MIRROR_SAMPLE_APP_ID = "2245261438923400";
#else
static const char* MIRROR_SAMPLE_APP_ID = "2227339760645525";
#endif

static const glm::vec3 LightPosition{0.0f, 2.2f, -1.0f};

#ifdef ANDROID
Application::Application(HMDType hmdType, jobject activity, JNIEnv* jniEnv)
    : HmdType(hmdType), Activity(activity), JniEnv(jniEnv) {
  // Initialize platform SDK
  PlatformInitializeRequest =
      ovr_PlatformInitializeAndroidAsynchronous(MIRROR_SAMPLE_APP_ID, activity, jniEnv);

  CommonInit("avatar_editor_environment.ktx", "avatar_editor_environment_128.ktx");
}
#else
Application::Application(HMDType hmdType) : HmdType(hmdType) {
  // Initialize platform SDK
  ovrPlatformInitializeResult platResult;
  PlatformInitializeRequest =
      ovr_PlatformInitializeWindowsAsynchronous(MIRROR_SAMPLE_APP_ID, &platResult);
  assert(platResult == ovrPlatformInitialize_Success);

  CommonInit(nullptr, "avatar_editor_environment_128.ktx");
}
#endif

Application::~Application() {
  Body.reset();
  Controllers.reset();
  Hands.reset();
  if (Avatar) {
    ovrAvatar_Destroy(Avatar);
  }
  if (AppState >= ovrAppState::RetrievingAvatarSpec) {
    ovrAvatar_Shutdown();
  }

  AudioEngineShutdown();
}

void Application::CommonInit(const char* environmentMap, const char* reflectionMap) {
  SetDebugCallback(
      [](GLenum source,
         GLenum type,
         GLuint id,
         GLenum severity,
         GLsizei length,
         const GLchar* message,
         const void* usrPtr) {
        OVR_LOG(
            "%d %d %d %d %.*s\n",
            (int)source,
            (int)type,
            (int)id,
            (int)severity,
            (int)length,
            message);
      },
      nullptr);

  // Unfortunately cube map timewarp layers aren't working correctly with Oculus OpenGL SDK.
  // So PC won't show the environment, will still show reflections though.
  if (environmentMap) {
    EnvTextureMap = CreateMemoryMap(environmentMap);
  }

  auto reflMap = CreateMemoryMap(reflectionMap);
  EnvReflectionMap = LoadKTXTexture((const uint8_t*)reflMap->data(), reflMap->size());

  SharedParams = ovrGlBuffer(ovrGlBufferType_t::GLBUFFER_TYPE_UNIFORM, sizeof(ovrSharedParams));
}

ovrUtilities::ovrTextureInfo Application::GetEnvTextureInfo() {
  return GetKTXTextureInfo((const uint8_t*)EnvTextureMap->data(), EnvTextureMap->size());
}

void Application::LoadEnvTexture(uint32_t texId, const ovrUtilities::ovrTextureInfo& texInfo) {
  glBindTexture(GL_TEXTURE_CUBE_MAP, texId);
  LoadKTXTexture((const uint8_t*)EnvTextureMap->data(), EnvTextureMap->size(), texInfo);
  glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
}

void Application::Frame(FrameParams& params) {
  switch (AppState) {
    case ovrAppState::PlatformInitializing:
      PlatformInitializing();
      break;
    case ovrAppState::CheckingEntitlement:
      CheckingEntitlement();
      break;
    case ovrAppState::RetrievingAvatarSpec:
      RetrievingAvatarSpec();
      break;
    case ovrAppState::RetrievingAssets:
      RetrievingAssets();
      break;
    case ovrAppState::Running:
      Running(params);
      break;
    default:
      break;
  }
}

template <typename CallableT>
void Application::ProccessPlatformSDKMsgs(const CallableT& callable) {
  ovrMessage* message = nullptr;
  while ((message = ovr_PopMessage()) != nullptr) {
    if (ovr_Message_IsError(message)) {
      ovrErrorHandle errorHandle = ovr_Message_GetError(message);
      int errorCode = ovr_Error_GetCode(errorHandle);
      int httpError = ovr_Error_GetHttpCode(errorHandle);
      OVR_WARN(
          "%d Error Code : %d ErrorMessage : %s ",
          httpError,
          errorCode,
          ovr_Error_GetMessage(errorHandle));
    } else {
      ovrRequest rqst = ovr_Message_GetRequestID(message);
      callable(rqst, message);
    }
    ovr_FreeMessage(message);
  }
}

template <typename CallableT>
void Application::ProccessAvatarSDKMsgs(const CallableT& callable) {
  ovrAvatarMessage* message = nullptr;
  while ((message = ovrAvatarMessage_Pop()) != nullptr) {
    ovrAvatarMessageType type = ovrAvatarMessage_GetType(message);
    callable(type, message);
    ovrAvatarMessage_Free(message);
  }
}

void Application::PlatformInitializing() {
  ProccessPlatformSDKMsgs([this](ovrRequest request, ovrMessage* message) {
    ovrRequest rqst = ovr_Message_GetRequestID(message);
    if (rqst == PlatformInitializeRequest) {
      OVR_LOG("platform initialized");

      PlatformInitializeRequest = 0;
      EntitlementRequest = ovr_Entitlement_GetIsViewerEntitled();

      AppState = ovrAppState::CheckingEntitlement;
    } else {
      OVR_LOG("Unexpected platform msg, was waiting for platform to be initialized");
    }
  });
}

void Application::CheckingEntitlement() {
  ProccessPlatformSDKMsgs([this](ovrRequest request, ovrMessage* message) {
    ovrRequest rqst = ovr_Message_GetRequestID(message);
    if (rqst == EntitlementRequest) {
      if (ovr_Message_IsError(message)) {
        OVR_FAIL("Entitlement Check Failed");
      }
      OVR_LOG("Entitlement check passed");
      // We are entitled!
      EntitlementRequest = 0;

      UserID = ovr_GetLoggedInUserID();
      // OVR_LOG("UserID: %" PRIu64, UserID);

      AudioEngineInitialize([](short*) {}, [this](short* buffer) { MicCallback(buffer); });

      // Initialize Avatar SDK
#ifdef ANDROID
      ovrAvatar_InitializeAndroid(MIRROR_SAMPLE_APP_ID, Activity, JniEnv);
#else
      ovrAvatar_Initialize(MIRROR_SAMPLE_APP_ID);
#endif

      // Now request our avatar specification
      auto requestSpec = ovrAvatarSpecificationRequest_Create(UserID);
      // For performance reasons, you should always use combined meshes.
      ovrAvatarSpecificationRequest_SetCombineMeshes(requestSpec, true);
      ovrAvatarSpecificationRequest_SetLookAndFeelVersion(
          requestSpec, ovrAvatarLookAndFeelVersion_Two);
      ovrAvatarSpecificationRequest_SetExpressiveFlag(requestSpec, true);
      ovrAvatar_RequestAvatarSpecificationFromSpecRequest(requestSpec);
      ovrAvatarSpecificationRequest_Destroy(requestSpec);

      AppState = ovrAppState::RetrievingAvatarSpec;
    } else {
      OVR_LOG("Unexpected platform msg, was waiting for platform to be initialized");
    }
  });
}

void Application::RetrievingAvatarSpec() {
  ProccessAvatarSDKMsgs([this](ovrAvatarMessageType type, ovrAvatarMessage* message) {
    switch (type) {
      case ovrAvatarMessageType_AvatarSpecification: {
        const ovrAvatarMessage_AvatarSpecification* spec =
            ovrAvatarMessage_GetAvatarSpecification(message);
        assert(spec->oculusUserID == UserID);

        // We have our avatars specification, so we can create it now. Can't render the avatar yet
        // though. Need to download all of the needed meshes and textures still.
        Avatar = ovrAvatar_Create(
            spec->avatarSpec,
            (ovrAvatarCapabilities)(
                ovrAvatarCapabilities::ovrAvatarCapability_Body |
                ovrAvatarCapabilities::ovrAvatarCapability_Expressive |
                ovrAvatarCapabilities::ovrAvatarCapability_Hands));

        ovrAvatar_SetLeftControllerVisibility(Avatar, true);
        ovrAvatar_SetLeftHandVisibility(Avatar, true);
        ovrAvatar_SetRightControllerVisibility(Avatar, true);
        ovrAvatar_SetRightHandVisibility(Avatar, true);

        Body = make_unique<ovrAvatarBody>(Avatar);
        Controllers = make_unique<ovrAvatarControllers>(Avatar, HmdType == HMDType::Rift);
        Hands = make_unique<ovrAvatarHands>(Avatar);

        AppState = ovrAppState::RetrievingAssets;
      } break;
      case ovrAvatarMessageType_AssetLoaded:
        OVR_WARN("Shouldn't be any assets loaded at this time");
        break;
      default:
        break;
    }
  });
}

void Application::RetrievingAssets() {
  ProccessAvatarSDKMsgs([this](ovrAvatarMessageType type, ovrAvatarMessage* message) {
    switch (type) {
      case ovrAvatarMessageType_AvatarSpecification:
        OVR_WARN("Shouldn't be any specs loaded at this time");
        break;
      case ovrAvatarMessageType_AssetLoaded: {
        const ovrAvatarMessage_AssetLoaded* assetMsg = ovrAvatarMessage_GetAssetLoaded(message);
        ovrAvatarAssetType type = ovrAvatarAsset_GetType(assetMsg->asset);
        // Combined Mesh will always be the body mesh. Controllers and hands are always regular
        // meshes
        if (type == ovrAvatarAssetType::ovrAvatarAssetType_CombinedMesh) {
          // For combined meshes, we dont use the assetId's. Just get the avatar
          // For our simple example, we only have a single avatar, so just assert that this mesh is
          // ours.
          [[maybe_unused]] const ovrAvatar* avatar = ovrAvatarAsset_GetAvatar(assetMsg->asset);
          assert(avatar == Avatar);
          Body->LoadMesh(assetMsg->asset);
        } else if (type == ovrAvatarAssetType::ovrAvatarAssetType_Mesh) {
          Controllers->LoadMesh(assetMsg->assetID, assetMsg->asset);
          Hands->LoadMesh(assetMsg->assetID, assetMsg->asset);
        } else if (type == ovrAvatarAssetType::ovrAvatarAssetType_Texture) {
          const ovrAvatarTextureAssetData* tex = ovrAvatarAsset_GetTextureData(assetMsg->asset);
          Body->LoadTexture(assetMsg->assetID, tex);
          Controllers->LoadTexture(assetMsg->assetID, tex);
          Hands->LoadTexture(assetMsg->assetID, tex);
        }

        if (Body->AllAssetsLoaded() && Controllers->AllAssetsLoaded() && Hands->AllAssetsLoaded()) {
          OVR_LOG("All assets loaded");
          AppState = ovrAppState::Running;
        }
      } break;
      default:
        break;
    }
  });
}

void Application::Running(FrameParams& params) {
  // On PC can sometimes have fames back to back for some reason(rare), Avatar SDK doesnt like that.
  // can just render the same thing again anyway, so do nothing here
  if (params.DeltaSeconds < 0.00005) {
    return;
  }

  // Render our own avatar moved in front of us, and "mirrored"
  const glm::vec3 MirrorOffset{0.0f, 0.0f, -1.0f};
  const glm::quat MirrorRotation{glm::vec3(0.0f, glm::radians(180.0f), 0.0f)}; // from euler angles

  // Handle microphone audio
  MicAudio([&](auto& micAudio) {
    if (!micAudio.empty()) {
      LipSync.ProcessAudio(data(micAudio), (uint32_t)micAudio.size());
      micAudio.clear();
    }
  });

  // Update our shared uniform Buffer
  ovrSharedParams* sharedParams = SharedParams.Map<ovrSharedParams>();
  sharedParams->ViewPosition = params.EyePos;
  sharedParams->LightPosition = LightPosition;
  SharedParams.Unmap();

  // Update our SDK Avatar
  glm::vec3 pos = params.EyePos + MirrorOffset;
  glm::quat rot = params.EyeRot;
  rot.x *= -1.0f; // mirror pitch
  rot.y *= -1.0f; // mirror yaw
  rot = rot * MirrorRotation;
  rot = glm::normalize(rot);
  ovrAvatarTransform bodyTransform;
  bodyTransform.orientation = {rot.x, rot.y, rot.z, rot.w};
  bodyTransform.position = {pos.x, pos.y, pos.z};
  bodyTransform.scale = {1.0f, 1.0f, 1.0f};

  ovrAvatarPose_UpdateBody(Avatar, bodyTransform);

  ovrAvatarHandInputState hState[2];
  for (uint32_t hand = 0; hand < 2; ++hand) {
    hState[hand].isActive = params.Hands[hand].isActive;
    if (params.Hands[hand].isActive) {
      glm::vec3 pos = params.Hands[hand].Position;
      pos.x *= -1.0f;
      pos.z *= -1.0f;
      pos += MirrorOffset;
      glm::quat rot = params.Hands[hand].Rotation;
      rot.x *= -1.0f; // mirror yaw
      rot.z *= -1.0f; // mirror roll
      rot = rot * MirrorRotation;
      rot = glm::normalize(rot);

      ovrAvatarTransform handTransform;
      handTransform.orientation = {rot.x, rot.y, rot.z, rot.w};
      handTransform.position = {pos.x, pos.y, pos.z};
      handTransform.scale = {1.0f, 1.0f, 1.0f};

      hState[hand].transform = handTransform;

      hState[hand].joystickX = params.Hands[hand].JoystickX;
      hState[hand].joystickY = params.Hands[hand].JoystickY;
      hState[hand].handTrigger = params.Hands[hand].HandTrigger;
      hState[hand].indexTrigger = params.Hands[hand].IndexTrigger;

      hState[hand].buttonMask = params.Hands[hand].ButtonMask;
      hState[hand].touchMask = params.Hands[hand].TouchMask;
    }
  }

  ovrAvatarControllerType controllerType =
      (HmdType == HMDType::Rift ? ovrAvatarControllerType::ovrAvatarControllerType_Touch
                                : ovrAvatarControllerType::ovrAvatarControllerType_Quest);
  ovrAvatarPose_UpdateHandsWithType(Avatar, hState[0], hState[1], controllerType);

  auto visemes = LipSync.GetVisemes();
  if (!visemes.empty()) {
    ovrAvatarVisemes sdkVisemes;
    sdkVisemes.visemeParamCount = (uint32_t)size(visemes);
    memcpy(sdkVisemes.visemeParams, data(visemes), sizeof(float) * size(visemes));
    ovrAvatar_SetVisemes(Avatar, &sdkVisemes);
  }

  ovrAvatarPose_Finalize(Avatar, params.DeltaSeconds);

  Body->Frame();
  Controllers->Frame();
  Hands->Frame();
}

void Application::MicCallback(short* buffer) {
  MicAudio(
      [buffer](auto& micAudio) { copy(buffer, buffer + MicBufferSize, back_inserter(micAudio)); });
}

void Application::Render(uint32_t eye, ovrUtilities::ovrGlBuffer& sceneMatrices) {
  if (AppState == ovrAppState::Running) {
    glDepthMask(GL_TRUE);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glFrontFace(GL_CCW);

    Body->Render(eye, SharedParams, EnvReflectionMap, sceneMatrices);
    Controllers->Render(eye, SharedParams, EnvReflectionMap, sceneMatrices);
    Hands->Render(eye, SharedParams, sceneMatrices);
  }
}
