#include "AvatarBody.h"
#include "Shaders.h"
#include "TextureLoader.h"
#include "TypeConversions.h"
#include "Utilities/GlProgram.h"
#include "Utilities/Log.h"
#include "VertexBufferWriter.h"

#include <algorithm>
#include <cassert>

#include "glm/gtc/packing.hpp"
#include "glm/gtc/type_ptr.hpp"
#include "glm/gtx/quaternion.hpp"
#include "glm/gtx/transform.hpp"
#include "glm/vec4.hpp"

using namespace std;
using namespace ovrUtilities;
using namespace glm;

namespace {
// Shader takes blends in groups of 4(for std140 reasons)
constexpr int BlendGroupSize = 4;
constexpr int MaxBlendParams = 16;
static_assert(
    MaxBlendParams % BlendGroupSize == 0,
    "number of blend shape params must be a multiple of 4");
// Ignore blends shapes with weight below this. Arbitrary. A large epsilon. Arbitrary, drop any that
// would not be noticeable. 0.1% should be effectively 0.
constexpr float MinBlendWeight = 0.001f;

// head, cloths, eyewear, hair, beard.
constexpr float DiffuseIntensity[ovrAvatarBodyPartType_Count] = {0.3f, 0.1f, 0.0f, 0.15f, 0.15f};
constexpr float ReflectionIntensity[ovrAvatarBodyPartType_Count] = {0.0f, 0.3f, 0.4f, 0.0f, 0.0f};

#pragma pack(push, 1)
// std140 layout, must match shader
struct MatVertParameters {
  ivec4 BSConstants; // x has size of a blend shape(in verts, i.e. 3 texels). y has number blend
                     // shapes used, always a multiple of 4
  ivec4 BSIndices[MaxBlendParams / BlendGroupSize]; // which blend shapes to blend
  vec4 BSWeights[MaxBlendParams / BlendGroupSize]; // blend shape weights
  mat4 Bones[9];
};

struct MatFragParameters {
  vec4 AlbedoTint[ovrAvatarBodyPartType_Count]; // One per renderpart.
  vec4 Intensity[ovrAvatarBodyPartType_Count]; // Diffuse in r, Reflection in g, brow in b, lip
                                               // smoothness in a
  vec4 Colors[7]; // Expressive Colors
};
#pragma pack(pop)

} // namespace

ovrAvatarBody::ovrAvatarBody(ovrAvatar* avatar) {
  BodyComp = ovrAvatarPose_GetBodyComponent(avatar);
  const ovrAvatarComponent* rcomp = BodyComp->renderComponent;

  ExpressiveParams = ovrAvatar_GetExpressiveParameters(avatar);

  MaterialFragmentBuffer =
      ovrGlBuffer(ovrGlBufferType_t::GLBUFFER_TYPE_UNIFORM, sizeof(MatFragParameters));
  MatFragParameters* matFragBuf = MaterialFragmentBuffer.Map<MatFragParameters>();

  TexturesToLoad.resize(ovrAvatarBodyPartType::ovrAvatarBodyPartType_Count);
  // Note, all of these renderParts will be replaced by a single renderPart after the combined mesh
  // has been loaded So you can not cache much here. Will need to go through the renderPart list
  // again later, on first Frame call, Mostly so we can get the bones and blend shape weights, which
  // are different for the combined renderPart.
  for (uint32_t rendIndex = 0; rendIndex < rcomp->renderPartCount; ++rendIndex) {
    const ovrAvatarRenderPart* rendPart = rcomp->renderParts[rendIndex];
    ovrAvatarRenderPartType type = ovrAvatarRenderPart_GetType(rendPart);
    // should always be a PBS V2, rest of code will assume that.
    assert(type == ovrAvatarRenderPartType_SkinnedMeshRenderPBS_V2);

    const ovrAvatarRenderPart_SkinnedMeshRenderPBS_V2* mesh =
        ovrAvatarRenderPart_GetSkinnedMeshRenderPBSV2(rendPart);
    assert(mesh->meshAssetID != 0);
    // Start fetching mesh asset(from cache, or downloaded. May take a while)
    ovrAvatarAsset_BeginLoading(mesh->meshAssetID);

    ovrAvatarBodyPartType renderType = ovrAvatarRenderPart_GetBodyPartType(avatar, rendPart);
    assert(renderType != ovrAvatarBodyPartType::ovrAvatarBodyPartType_Count);

    const ovrAvatarPBSMaterialState& mat = mesh->materialState;
    TextureToLoad textures{mat.albedoTextureID, mat.normalTextureID, mat.metallicnessTextureID};
    TexturesToLoad[renderType] = textures;
    // Start fetching texture asset(from cache, or downloaded. May take a while)
    ovrAvatarAsset_BeginLoading(mat.albedoTextureID);
    ovrAvatarAsset_BeginLoading(mat.normalTextureID);
    ovrAvatarAsset_BeginLoading(mat.metallicnessTextureID);

    matFragBuf->AlbedoTint[renderType] = make_vec4(&mat.albedoMultiplier.x);

    matFragBuf->Intensity[renderType].x = DiffuseIntensity[renderType];
    matFragBuf->Intensity[renderType].y = ReflectionIntensity[renderType];
    // Note these 2 are only used for body part, just write same values into all fields.
    matFragBuf->Intensity[renderType].z = ExpressiveParams.browLashIntensity;
    matFragBuf->Intensity[renderType].w = ExpressiveParams.lipSmoothness;
  }

  matFragBuf->Colors[0] = make_vec4(&ExpressiveParams.browColor.x);
  matFragBuf->Colors[1] = make_vec4(&ExpressiveParams.teethColor.x);
  matFragBuf->Colors[2] = make_vec4(&ExpressiveParams.gumColor.x);
  matFragBuf->Colors[3] = make_vec4(&ExpressiveParams.lipColor.x);
  matFragBuf->Colors[4] = make_vec4(&ExpressiveParams.lashColor.x);
  matFragBuf->Colors[5] = make_vec4(&ExpressiveParams.scleraColor.x);
  matFragBuf->Colors[6] = make_vec4(&ExpressiveParams.irisColor.x);

  MaterialFragmentBuffer.Unmap();

  // Create this buffer here, but nothing to write to it at this time, it will be updated every
  // frame.
  MaterialVertexBuffer =
      ovrGlBuffer(ovrGlBufferType_t::GLBUFFER_TYPE_UNIFORM, sizeof(MatVertParameters));
}

bool ovrAvatarBody::AllAssetsLoaded() const {
  return MeshLoaded && TexturesToLoad.empty();
}

void ovrAvatarBody::LoadMesh(const ovrAvatarAsset* asset) {
  OVR_LOG("ovrAvatarBody::LoadMesh");

  const ovrAvatarMeshAssetDataV2* mesh = ovrAvatarAsset_GetCombinedMeshData(asset);

  Mesh = ovrMesh{asset, mesh};

  // And blend shapes too.
  uint32_t blendShapeCount = ovrAvatarAsset_GetMeshBlendShapeCount(asset);
  assert(blendShapeCount != 0);

  const ovrAvatarBlendVertex* blendShapes = ovrAvatarAsset_GetMeshBlendShapeVertices(asset);

  // Blend shapes are large. cut size in half by using half floats. each uint64 is 4 half floats
#pragma pack(push, 1)
  struct BlendShapeVertex {
    uint64 Position; // w unused, just for padding
    uint64 Normal; // w unused
    uint64 Tangent; // w unused
  };
#pragma pack(pop)

  // prep data first
  std::vector<BlendShapeVertex> BlendShapes;
  uint32_t vertIndex = 0;
  for (uint32_t shape = 0; shape < blendShapeCount; ++shape) {
    for (uint32_t vertex = 0; vertex < mesh->vertexCount; ++vertex) {
      // make a copy of the vert, not safe to access directly, alignment bug in Avatar SDK
      ovrAvatarBlendVertex blendVert;
      memcpy((void*)&blendVert, (void*)(blendShapes + vertIndex), sizeof(ovrAvatarBlendVertex));

      BlendShapes.emplace_back();
      auto& shape = BlendShapes.back();
      shape.Position = packHalf4x16(vec4{blendVert.x, blendVert.y, blendVert.z, 0.0f});
      shape.Normal = packHalf4x16(vec4{blendVert.nx, blendVert.ny, blendVert.nz, 0.0f});
      shape.Tangent = packHalf4x16(vec4{blendVert.tx, blendVert.ty, blendVert.tz, 0.0f});

      ++vertIndex;
    }
  }

  BlendShapeSize = mesh->vertexCount;

  // now generate the actual texture buffer
  BShapeTextureBuffer = ovrGlBuffer(
      ovrGlBufferType_t::GLBUFFER_TYPE_TEXTURE_BUFFER,
      size(BlendShapes) * sizeof(BlendShapeVertex));
  assert(BShapeTextureBuffer.IsValid());
  BShapeTexture = ovrGlTexture(BShapeTextureBuffer, GL_RGBA16F);
  assert(BShapeTexture.IsValid());

  BlendShapeVertex* bshapeBuffer = BShapeTextureBuffer.Map<BlendShapeVertex>();
  memcpy(bshapeBuffer, data(BlendShapes), size(BlendShapes) * sizeof(BlendShapeVertex));
  BShapeTextureBuffer.Unmap();

  MeshLoaded = true;
}

void ovrAvatarBody::LoadTexture(ovrAvatarAssetID id, const ovrAvatarTextureAssetData* tex) {
  OVR_LOG("ovrAvatarBody::LoadTexture");

  for (uint32_t i = 0; i < TexturesToLoad.size(); ++i) {
    if (TexturesToLoad[i].Albedo == id) {
      // These checks are required because of some bad assets currently. These textures won't show
      // up.
      if (tex->format == ovrAvatarTextureFormat_DXT1) {
        OVR_WARN("body albedo texture %llu was DXT1, should always be DXT5", id);
      } else {
        ::LoadTexture(AlbedoTextures, i, tex, true);
      }
      TexturesToLoad[i].Albedo = 0;
    }
    if (TexturesToLoad[i].Normal == id) {
      if (tex->format == ovrAvatarTextureFormat_DXT5) {
        OVR_WARN("body normal texture %llu was DXT5, should always be DXT1", id);
      } else {
        ::LoadTexture(NormalTextures, i, tex, false);
      }
      TexturesToLoad[i].Normal = 0;
    }
    if (TexturesToLoad[i].Metallic == id) {
      if (tex->format == ovrAvatarTextureFormat_DXT1) {
        OVR_WARN("body metallic texture %llu was DXT1, should always be DXT5", id);
      } else {
        ::LoadTexture(MetallicTextures, i, tex, false);
      }
      TexturesToLoad[i].Metallic = 0;
    }
  }

  // If All textures have been loaded, clear vector so we know we are done
  if (all_of(begin(TexturesToLoad), end(TexturesToLoad), [](auto& tex) {
        return tex.Albedo == 0 && tex.Normal == 0 && tex.Metallic == 0;
      })) {
    TexturesToLoad.clear();
  }
}

void ovrAvatarBody::Frame() {
  // First frame, cache the new combined render part
  if (!CombinedRenderPart) {
    assert(BodyComp);
    const ovrAvatarComponent* rcomp = BodyComp->renderComponent;

    // should be a single render part for the combined mesh.
    assert(rcomp->renderPartCount == 1);
    CombinedRenderPart = rcomp->renderParts[0];
    CombinedRenderComp = rcomp;
  }

  assert(CombinedRenderPart);
  const ovrAvatarBlendShapeParams* blendShapeParms =
      ovrAvatarSkinnedMeshRender_GetBlendShapeParams(CombinedRenderPart);

  // Dont send more blends than we need. We do send a multiple of 4 though.
  vector<pair<int32_t, float>> importantBlendParams;
  importantBlendParams.clear();
  for (int32_t param = 0; param < (int32_t)blendShapeParms->blendShapeParamCount; ++param) {
    importantBlendParams.emplace_back(param, blendShapeParms->blendShapeParams[param]);
  }
  // sort so most important blends first
  sort(
      begin(importantBlendParams),
      end(importantBlendParams),
      [](const auto& arg1, const auto& arg2) { return arg1.second > arg2.second; });
  // now remove least important if there are more than we support in the shader.
  if (importantBlendParams.size() > MaxBlendParams) {
    importantBlendParams.resize(MaxBlendParams);
  }
  // We must have a multiple of 4, so pad with 0's to next multiple.
  while ((importantBlendParams.size() % BlendGroupSize) != 0) {
    importantBlendParams.emplace_back(0, 0.0f);
  }
  // now remove any that are too small to matter, have to do in group of 4's
  while (!importantBlendParams.empty() &&
         (importantBlendParams.end() - BlendGroupSize)->second < MinBlendWeight) {
    importantBlendParams.erase(importantBlendParams.end() - 4, importantBlendParams.end());
  }

  MatVertParameters* matVertParms = MaterialVertexBuffer.Map<MatVertParameters>();

  for (int group = 0; group < (importantBlendParams.size() / BlendGroupSize); ++group) {
    matVertParms->BSIndices[group] = {importantBlendParams[BlendGroupSize * group + 0].first,
                                      importantBlendParams[BlendGroupSize * group + 1].first,
                                      importantBlendParams[BlendGroupSize * group + 2].first,
                                      importantBlendParams[BlendGroupSize * group + 3].first};
    matVertParms->BSWeights[group] = {importantBlendParams[BlendGroupSize * group + 0].second,
                                      importantBlendParams[BlendGroupSize * group + 1].second,
                                      importantBlendParams[BlendGroupSize * group + 2].second,
                                      importantBlendParams[BlendGroupSize * group + 3].second};
  }

  matVertParms->BSConstants.x = BlendShapeSize;
  matVertParms->BSConstants.y = (int)importantBlendParams.size() / BlendGroupSize;
  matVertParms->BSConstants.z = 0;
  matVertParms->BSConstants.w = 0;

  const ovrAvatarRenderPart_SkinnedMeshRenderPBS_V2* skinnedMesh =
      ovrAvatarRenderPart_GetSkinnedMeshRenderPBSV2(CombinedRenderPart);
  const ovrAvatarSkinnedMeshPose& skinnedPose = skinnedMesh->skinnedPose;
  //+1 to avoid a branch, if bone doesn't have parent(index will be -1) then use 0 index which will
  // be identity.
  mat4 worldPose[MAX_JOINTS + 1];
  worldPose[0] = glm::identity<mat4>();
  const int* jointParents = skinnedPose.jointParents;
  const ovrAvatarTransform* jointTransforms = skinnedPose.jointTransform;
  assert(skinnedPose.jointCount <= MAX_JOINTS);

  for (uint32_t i = 0; i < skinnedPose.jointCount; i++) {
    mat4 local = Convert(jointTransforms[i]);

    const int parentIndex = jointParents[i];
    assert(parentIndex >= -1);
    worldPose[i + 1] = worldPose[parentIndex + 1] * local;
    matVertParms->Bones[i] = worldPose[i + 1] * Mesh.GetInverseBindPose()[i];
  }

  MaterialVertexBuffer.Unmap();

  WorldTransform = Convert(CombinedRenderComp->transform) * Convert(skinnedMesh->localTransform);
}

void ovrAvatarBody::Render(
    uint32_t eye,
    ovrUtilities::ovrGlBuffer& sharedParams,
    ovrUtilities::ovrGlTexture& envMap,
    ovrUtilities::ovrGlBuffer& sceneMatrices) {
  const ovrGlProgram& program = GetBodyProgram();

  // Setup
  glUseProgram(program.Program);

  sceneMatrices.Bind(program.SceneMatrices.Binding);
  sharedParams.Bind(program.Uniforms[(int)UniformIndexBody::SharedParms].Binding);
  MaterialVertexBuffer.Bind(program.Uniforms[(int)UniformIndexBody::VertMaterials].Binding);
  MaterialFragmentBuffer.Bind(program.Uniforms[(int)UniformIndexBody::FragMaterials].Binding);

  AlbedoTextures.Activate(program.Uniforms[(int)UniformIndexBody::Albedo].Binding);
  NormalTextures.Activate(program.Uniforms[(int)UniformIndexBody::Normal].Binding);
  MetallicTextures.Activate(program.Uniforms[(int)UniformIndexBody::Metallic].Binding);
  BShapeTexture.Activate(program.Uniforms[(int)UniformIndexBody::BlendShapes].Binding);
  envMap.Activate(program.Uniforms[(int)UniformIndexBody::EnvironmentMap].Binding);

  glEnable(GL_BLEND);
  glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
  glBlendEquationSeparate(GL_FUNC_ADD, GL_FUNC_ADD);

  if (program.ViewID.Location >= 0) // not defined when multiview enabled
  {
    glUniform1i(program.ViewID.Location, eye);
  }
  glUniformMatrix4fv(program.ModelMatrix.Binding, 1, GL_FALSE, (float*)&WorldTransform);

  // Draw
  glBindVertexArray(Mesh.GetVertexArray());
  glDrawElements(GL_TRIANGLES, Mesh.GetIndexCount(), GL_UNSIGNED_SHORT, nullptr);

  // cleanup
  AlbedoTextures.DeActivate(program.Uniforms[(int)UniformIndexBody::Albedo].Binding);
  NormalTextures.DeActivate(program.Uniforms[(int)UniformIndexBody::Normal].Binding);
  MetallicTextures.DeActivate(program.Uniforms[(int)UniformIndexBody::Metallic].Binding);
  BShapeTexture.DeActivate(program.Uniforms[(int)UniformIndexBody::BlendShapes].Binding);
  envMap.DeActivate(program.Uniforms[(int)UniformIndexBody::EnvironmentMap].Binding);

  sceneMatrices.Unbind(program.SceneMatrices.Binding);
  sharedParams.Unbind(program.Uniforms[(int)UniformIndexBody::SharedParms].Binding);
  MaterialVertexBuffer.Unbind(program.Uniforms[(int)UniformIndexBody::VertMaterials].Binding);
  MaterialFragmentBuffer.Unbind(program.Uniforms[(int)UniformIndexBody::FragMaterials].Binding);

  glBindVertexArray(0);
  glUseProgram(0);
}
