#include "AvatarControllers.h"

#include "Shaders.h"
#include "TextureLoader.h"
#include "TypeConversions.h"
#include "Utilities/GlProgram.h"
#include "Utilities/Log.h"

#include <algorithm>
#include <cassert>

#include "glm/gtc/packing.hpp"
#include "glm/gtc/type_ptr.hpp"
#include "glm/gtx/quaternion.hpp"
#include "glm/gtx/transform.hpp"
#include "glm/vec4.hpp"

using namespace std;
using namespace ovrUtilities;
using namespace glm;

namespace {
constexpr uint32_t ControllerBoneCount = 8;
enum ControllerRenderPartIndex {
  Rift = 0, // A V1 PBS render part
  Quest = 1, // A V2 PBS render part, also used for RiftS
  Gear = 2, // unused in this sample
  Go = 3, // unused in this sample
};
#pragma pack(push, 1)
// std140 layout, must match shader
struct MatVertParameters {
  mat4 Bones[ControllerBoneCount];
};
#pragma pack(pop)
static_assert(sizeof(MatVertParameters) % 256 == 0);
} // namespace

ovrAvatarControllers::ovrAvatarControllers(ovrAvatar* avatar, bool cv1Controller) {
  const ovrAvatarControllerComponent* contComps[] = {nullptr, nullptr};
  contComps[0] = ovrAvatarPose_GetLeftControllerComponent(avatar);
  contComps[1] = ovrAvatarPose_GetRightControllerComponent(avatar);

  TexturesToLoad.resize(2);
  MeshesToLoad.resize(2);
  for (uint32_t hand = 0; hand < 2; ++hand) {
    const ovrAvatarComponent* rcomp = contComps[hand]->renderComponent;
    // Default API is to load all render parts here, however, that unnecessarily loads meshes and
    // textures that will never be used. Each render part is for a single controller type. So
    // instead just ignore the render parts we dont care about.
    ControllerRenderPartIndex rendIndex = cv1Controller ? Rift : Quest;
    assert(rcomp->renderPartCount > rendIndex);

    const ovrAvatarRenderPart* rendPart = rcomp->renderParts[rendIndex];
    ovrAvatarRenderPartType type = ovrAvatarRenderPart_GetType(rendPart);
    // type should always be a PBS "V1" for RiftCV1 and PBS V2 for Quest/RiftS
    assert(
        type == ovrAvatarRenderPartType_SkinnedMeshRenderPBS ||
        type == ovrAvatarRenderPartType_SkinnedMeshRenderPBS_V2);
    if (type == ovrAvatarRenderPartType_SkinnedMeshRenderPBS) {
      PBSV2 = false;
      const ovrAvatarRenderPart_SkinnedMeshRenderPBS* mesh =
          ovrAvatarRenderPart_GetSkinnedMeshRenderPBS(rendPart);
      assert(mesh->meshAssetID != 0);
      // Start fetching mesh asset(from cache, or downloaded. May take a while)
      ovrAvatarAsset_BeginLoading(mesh->meshAssetID);
      MeshesToLoad[hand] = mesh->meshAssetID;

      TextureToLoad textures{mesh->albedoTextureAssetID, 0, mesh->surfaceTextureAssetID};
      TexturesToLoad[hand] = textures;
      // Start fetching texture asset(from cache, or downloaded. May take a while)
      ovrAvatarAsset_BeginLoading(mesh->albedoTextureAssetID);
      ovrAvatarAsset_BeginLoading(mesh->surfaceTextureAssetID);
    } else {
      // ovrAvatarRenderPartType_SkinnedMeshRenderPBS_V2
      PBSV2 = true;
      const ovrAvatarRenderPart_SkinnedMeshRenderPBS_V2* mesh =
          ovrAvatarRenderPart_GetSkinnedMeshRenderPBSV2(rendPart);
      assert(mesh->meshAssetID != 0);
      // Start fetching mesh asset(from cache, or downloaded. May take a while)
      ovrAvatarAsset_BeginLoading(mesh->meshAssetID);
      MeshesToLoad[hand] = mesh->meshAssetID;

      const ovrAvatarPBSMaterialState& mat = mesh->materialState;
      TextureToLoad textures{mat.albedoTextureID, mat.normalTextureID, mat.metallicnessTextureID};
      TexturesToLoad[hand] = textures;
      // Start fetching texture asset(from cache, or downloaded. May take a while)
      ovrAvatarAsset_BeginLoading(mat.albedoTextureID);
      ovrAvatarAsset_BeginLoading(mat.normalTextureID);
      ovrAvatarAsset_BeginLoading(mat.metallicnessTextureID);
    }

    RenderParts[hand] = rendPart;
    RenderComps[hand] = rcomp;
  }

  // Create this buffer here, but nothing to write to it at this time, it will be updated every
  // frame.
  MaterialVertexBuffer =
      ovrGlBuffer(ovrGlBufferType_t::GLBUFFER_TYPE_UNIFORM, sizeof(MatVertParameters) * 2);
}

bool ovrAvatarControllers::AllAssetsLoaded() const {
  return MeshesToLoad.empty() && TexturesToLoad.empty();
}

void ovrAvatarControllers::LoadMesh(ovrAvatarAssetID id, const ovrAvatarAsset* asset) {
  OVR_LOG("ovrAvatarControllers::LoadMesh");

  int32_t hand = -1;
  if (MeshesToLoad.empty()) {
    return;
  }

  if (MeshesToLoad[0] == id) {
    hand = 0;
    MeshesToLoad[0] = 0;
  } else if (MeshesToLoad[1] == id) {
    hand = 1;
    MeshesToLoad[1] = 0;
  } else {
    return; // not our mesh
  }

  const ovrAvatarMeshAssetData* mesh = ovrAvatarAsset_GetMeshData(asset);

  Meshes[hand] = ovrMesh{asset, mesh};

  // If All meshes have been loaded, clear vector so we know we are done
  if (all_of(begin(MeshesToLoad), end(MeshesToLoad), [](auto& tex) { return tex == 0; })) {
    MeshesToLoad.clear();
  }
}

void ovrAvatarControllers::LoadTexture(ovrAvatarAssetID id, const ovrAvatarTextureAssetData* tex) {
  OVR_LOG("ovrAvatarControllers::LoadTexture");

  for (uint32_t i = 0; i < TexturesToLoad.size(); ++i) {
    if (TexturesToLoad[i].Albedo == id) {
      ::LoadTexture(AlbedoTextures, i, tex, true);
      TexturesToLoad[i].Albedo = 0;
    }
    if (TexturesToLoad[i].Normal == id) {
      if (tex->format == ovrAvatarTextureFormat_DXT5) {
        OVR_WARN("controller normal texture %llu was DXT5, should always be DXT1", id);
      } else {
        ::LoadTexture(NormalTextures, i, tex, false);
      }
      TexturesToLoad[i].Normal = 0;
    }
    if (TexturesToLoad[i].Metallic == id) {
      if (tex->format == ovrAvatarTextureFormat_DXT1) {
        OVR_WARN("controller metallic texture %llu was DXT1, should always be DXT5", id);
      } else {
        ::LoadTexture(MetallicTextures, i, tex, false);
      }
      TexturesToLoad[i].Metallic = 0;
    }
  }

  // If All textures have been loaded, clear vector so we know we are done
  if (all_of(begin(TexturesToLoad), end(TexturesToLoad), [](auto& tex) {
        return tex.Albedo == 0 && tex.Metallic == 0;
      })) {
    TexturesToLoad.clear();
  }
}

void ovrAvatarControllers::Frame() {
  MatVertParameters* matVertParms = MaterialVertexBuffer.Map<MatVertParameters>();

  for (uint32_t hand = 0; hand < 2; ++hand) {
    mat4 localTransform;
    const ovrAvatarSkinnedMeshPose* skinnedPose = nullptr;
    if (PBSV2) {
      const ovrAvatarRenderPart_SkinnedMeshRenderPBS_V2* skinnedMesh =
          ovrAvatarRenderPart_GetSkinnedMeshRenderPBSV2(RenderParts[hand]);
      skinnedPose = &skinnedMesh->skinnedPose;
      localTransform = Convert(skinnedMesh->localTransform);
    } else {
      const ovrAvatarRenderPart_SkinnedMeshRenderPBS* skinnedMesh =
          ovrAvatarRenderPart_GetSkinnedMeshRenderPBS(RenderParts[hand]);
      skinnedPose = &skinnedMesh->skinnedPose;
      localTransform = Convert(skinnedMesh->localTransform);
    }

    //+1 to avoid a branch, if bone doesn't have parent(index will be -1) then use 0 index which
    // will
    // be identity.
    mat4 worldPose[MAX_JOINTS + 1];
    worldPose[0] = glm::identity<mat4>();
    const int* jointParents = skinnedPose->jointParents;
    const ovrAvatarTransform* jointTransforms = skinnedPose->jointTransform;
    assert(skinnedPose->jointCount <= MAX_JOINTS);

    for (uint32_t i = 0; i < skinnedPose->jointCount; i++) {
      mat4 local = Convert(jointTransforms[i]);

      const int parentIndex = jointParents[i];
      assert(parentIndex >= -1);
      worldPose[i + 1] = worldPose[parentIndex + 1] * local;
      matVertParms[hand].Bones[i] = worldPose[i + 1] * Meshes[hand].GetInverseBindPose()[i];
    }

    WorldTransforms[hand] = Convert(RenderComps[hand]->transform) * localTransform;
  }
  MaterialVertexBuffer.Unmap();
}

void ovrAvatarControllers::Render(
    uint32_t eye,
    ovrUtilities::ovrGlBuffer& sharedParams,
    ovrUtilities::ovrGlTexture& envMap,
    ovrUtilities::ovrGlBuffer& sceneMatrices) {
  const ovrGlProgram& program = PBSV2 ? GetControllerProgramV2() : GetControllerProgramV1();

  // Setup
  glUseProgram(program.Program);

  sceneMatrices.Bind(program.SceneMatrices.Binding);

  sharedParams.Bind(program.Uniforms[(int)UniformIndexController::SharedParms].Binding);
  AlbedoTextures.Activate(program.Uniforms[(int)UniformIndexController::Albedo].Binding);
  MetallicTextures.Activate(program.Uniforms[(int)UniformIndexController::Metallic].Binding);
  if (PBSV2) {
    NormalTextures.Activate(program.Uniforms[(int)UniformIndexController::Normal].Binding);
  }
  envMap.Activate(program.Uniforms[(int)UniformIndexController::EnvironmentMap].Binding);

  glDisable(GL_BLEND);

  if (program.ViewID.Location >= 0) // not defined when multiview enabled
  {
    glUniform1i(program.ViewID.Location, eye);
  }

  for (uint32_t hand = 0; hand < 2; ++hand) {
    MaterialVertexBuffer.BindRange(
        program.Uniforms[(int)UniformIndexController::VertMaterials].Binding,
        hand * sizeof(MatVertParameters),
        sizeof(MatVertParameters));

    glUniformMatrix4fv(program.ModelMatrix.Binding, 1, GL_FALSE, (float*)&WorldTransforms[hand]);
    glUniform1i(program.Uniforms[(int)UniformIndexController::Hand].Binding, hand);

    // Draw
    glBindVertexArray(Meshes[hand].GetVertexArray());
    glDrawElements(GL_TRIANGLES, Meshes[hand].GetIndexCount(), GL_UNSIGNED_SHORT, nullptr);
  }

  // cleanup
  AlbedoTextures.DeActivate(program.Uniforms[(int)UniformIndexController::Albedo].Binding);
  MetallicTextures.DeActivate(program.Uniforms[(int)UniformIndexController::Metallic].Binding);
  if (PBSV2) {
    NormalTextures.DeActivate(program.Uniforms[(int)UniformIndexController::Normal].Binding);
  }
  envMap.DeActivate(program.Uniforms[(int)UniformIndexController::EnvironmentMap].Binding);

  sceneMatrices.Unbind(program.SceneMatrices.Binding);
  sharedParams.Unbind(program.Uniforms[(int)UniformIndexController::SharedParms].Binding);
  MaterialVertexBuffer.Unbind(program.Uniforms[(int)UniformIndexController::VertMaterials].Binding);

  glBindVertexArray(0);
  glUseProgram(0);
}
