#include "AvatarHands.h"

#include "Shaders.h"
#include "TextureLoader.h"
#include "TypeConversions.h"
#include "Utilities/GlProgram.h"
#include "Utilities/Log.h"

#include <algorithm>
#include <cassert>

#include "glm/gtc/packing.hpp"
#include "glm/gtc/type_ptr.hpp"
#include "glm/gtx/quaternion.hpp"
#include "glm/gtx/transform.hpp"
#include "glm/vec4.hpp"

using namespace std;
using namespace ovrUtilities;
using namespace glm;

namespace {
constexpr uint32_t HandBoneCount = 28;

#pragma pack(push, 1)
// std140 layout, must match shader
struct MatVertParameters {
  mat4 Bones[HandBoneCount];
};
#pragma pack(pop)
static_assert(sizeof(MatVertParameters) % 256 == 0);
} // namespace

ovrAvatarHands::ovrAvatarHands(ovrAvatar* avatar) {
  const ovrAvatarHandComponent* handComps[] = {nullptr, nullptr};
  handComps[0] = ovrAvatarPose_GetLeftHandComponent(avatar);
  handComps[1] = ovrAvatarPose_GetRightHandComponent(avatar);

  TexturesToLoad.resize(2);
  MeshesToLoad.resize(2);
  for (uint32_t hand = 0; hand < 2; ++hand) {
    const ovrAvatarComponent* rcomp = handComps[hand]->renderComponent;
    assert(rcomp->renderPartCount == 1); // should only be one render component on a hand

    const ovrAvatarRenderPart* rendPart = rcomp->renderParts[0];
    ovrAvatarRenderPartType type = ovrAvatarRenderPart_GetType(rendPart);
    // should always be a PBS V2, rest of code will assume that.
    assert(type == ovrAvatarRenderPartType_SkinnedMeshRenderPBS_V2);

    const ovrAvatarRenderPart_SkinnedMeshRenderPBS_V2* mesh =
        ovrAvatarRenderPart_GetSkinnedMeshRenderPBSV2(rendPart);
    assert(mesh->meshAssetID != 0);
    // Start fetching mesh asset(from cache, or downloaded. May take a while)
    ovrAvatarAsset_BeginLoading(mesh->meshAssetID);
    MeshesToLoad[hand] = mesh->meshAssetID;

    const ovrAvatarPBSMaterialState& mat = mesh->materialState;
    TextureToLoad textures{mat.albedoTextureID, mat.normalTextureID};
    TexturesToLoad[hand] = textures;
    // Start fetching texture asset(from cache, or downloaded. May take a while)
    ovrAvatarAsset_BeginLoading(mat.albedoTextureID);
    ovrAvatarAsset_BeginLoading(mat.normalTextureID);

    AlbedoTints[hand] = make_vec4(&mat.albedoMultiplier.x);

    RenderParts[hand] = rendPart;
    RenderComps[hand] = rcomp;
  }

  // Create this buffer here, but nothing to write to it at this time, it will be updated every
  // frame.
  MaterialVertexBuffer =
      ovrGlBuffer(ovrGlBufferType_t::GLBUFFER_TYPE_UNIFORM, sizeof(MatVertParameters) * 2);
}

bool ovrAvatarHands::AllAssetsLoaded() const {
  return MeshesToLoad.empty() && TexturesToLoad.empty();
}

void ovrAvatarHands::LoadMesh(ovrAvatarAssetID id, const ovrAvatarAsset* asset) {
  OVR_LOG("ovrAvatarHands::LoadMesh");

  int32_t hand = -1;
  if (MeshesToLoad.empty()) {
    return;
  }

  if (MeshesToLoad[0] == id) {
    hand = 0;
    MeshesToLoad[0] = 0;
  } else if (MeshesToLoad[1] == id) {
    hand = 1;
    MeshesToLoad[1] = 0;
  } else {
    return; // not our mesh
  }

  const ovrAvatarMeshAssetData* mesh = ovrAvatarAsset_GetMeshData(asset);

  Meshes[hand] = ovrMesh{asset, mesh};

  // If All meshes have been loaded, clear vector so we know we are done
  if (all_of(begin(MeshesToLoad), end(MeshesToLoad), [](auto& tex) { return tex == 0; })) {
    MeshesToLoad.clear();
  }
}

void ovrAvatarHands::LoadTexture(ovrAvatarAssetID id, const ovrAvatarTextureAssetData* tex) {
  OVR_LOG("ovrAvatarHands::LoadTexture");

  for (uint32_t i = 0; i < TexturesToLoad.size(); ++i) {
    if (TexturesToLoad[i].Albedo == id) {
      // These checks are required because of some bad assets currently. These textures won't show
      // up.
      if (tex->format == ovrAvatarTextureFormat_DXT1) {
        OVR_WARN("hand albedo texture %llu was DXT1, should always be DXT5", id);
      } else {
        ::LoadTexture(AlbedoTextures, i, tex, true);
      }
      TexturesToLoad[i].Albedo = 0;
    }
    if (TexturesToLoad[i].Normal == id) {
      if (tex->format == ovrAvatarTextureFormat_DXT5) {
        OVR_WARN("hand normal texture %llu was DXT5, should always be DXT1", id);
      } else {
        ::LoadTexture(NormalTextures, i, tex, false);
      }
      TexturesToLoad[i].Normal = 0;
    }
  }

  // If All textures have been loaded, clear vector so we know we are done
  if (all_of(begin(TexturesToLoad), end(TexturesToLoad), [](auto& tex) {
        return tex.Albedo == 0 && tex.Normal == 0;
      })) {
    TexturesToLoad.clear();
  }
}

void ovrAvatarHands::Frame() {
  MatVertParameters* matVertParms = MaterialVertexBuffer.Map<MatVertParameters>();

  for (uint32_t hand = 0; hand < 2; ++hand) {
    const ovrAvatarRenderPart_SkinnedMeshRenderPBS_V2* skinnedMesh =
        ovrAvatarRenderPart_GetSkinnedMeshRenderPBSV2(RenderParts[hand]);
    const ovrAvatarSkinnedMeshPose& skinnedPose = skinnedMesh->skinnedPose;
    //+1 to avoid a branch, if bone doesn't have parent(index will be -1) then use 0 index which
    // will
    // be identity.
    mat4 worldPose[MAX_JOINTS + 1];
    worldPose[0] = glm::identity<mat4>();
    const int* jointParents = skinnedPose.jointParents;
    const ovrAvatarTransform* jointTransforms = skinnedPose.jointTransform;
    assert(skinnedPose.jointCount <= MAX_JOINTS);

    for (uint32_t i = 0; i < skinnedPose.jointCount; i++) {
      mat4 local = Convert(jointTransforms[i]);

      const int parentIndex = jointParents[i];
      assert(parentIndex >= -1);
      worldPose[i + 1] = worldPose[parentIndex + 1] * local;
      matVertParms[hand].Bones[i] = worldPose[i + 1] * Meshes[hand].GetInverseBindPose()[i];
    }

    WorldTransforms[hand] =
        Convert(RenderComps[hand]->transform) * Convert(skinnedMesh->localTransform);
  }
  MaterialVertexBuffer.Unmap();
}

void ovrAvatarHands::Render(
    uint32_t eye,
    ovrUtilities::ovrGlBuffer& sharedParams,
    ovrUtilities::ovrGlBuffer& sceneMatrices) {
  const ovrGlProgram& program = GetHandProgram();

  // Setup
  glUseProgram(program.Program);

  sceneMatrices.Bind(program.SceneMatrices.Binding);
  sharedParams.Bind(program.Uniforms[(int)UniformIndexHand::SharedParms].Binding);

  AlbedoTextures.Activate(program.Uniforms[(int)UniformIndexHand::Albedo].Binding);
  NormalTextures.Activate(program.Uniforms[(int)UniformIndexHand::Normal].Binding);

  glEnable(GL_BLEND);
  glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
  glBlendEquationSeparate(GL_FUNC_ADD, GL_FUNC_ADD);

  if (program.ViewID.Location >= 0) // not defined when multiview enabled
  {
    glUniform1i(program.ViewID.Location, eye);
  }

  for (uint32_t hand = 0; hand < 2; ++hand) {
    MaterialVertexBuffer.BindRange(
        program.Uniforms[(int)UniformIndexHand::VertMaterials].Binding,
        hand * sizeof(MatVertParameters),
        sizeof(MatVertParameters));

    glUniformMatrix4fv(program.ModelMatrix.Binding, 1, GL_FALSE, (float*)&WorldTransforms[hand]);
    glUniform4fv(
        program.Uniforms[(int)UniformIndexHand::AlbedoTint].Binding, 1, &AlbedoTints[hand].x);
    glUniform1i(program.Uniforms[(int)UniformIndexHand::Hand].Binding, hand);

    // Draw
    glBindVertexArray(Meshes[hand].GetVertexArray());
    glDrawElements(GL_TRIANGLES, Meshes[hand].GetIndexCount(), GL_UNSIGNED_SHORT, nullptr);
  }

  // cleanup
  AlbedoTextures.DeActivate(program.Uniforms[(int)UniformIndexBody::Albedo].Binding);
  NormalTextures.DeActivate(program.Uniforms[(int)UniformIndexHand::Normal].Binding);

  sceneMatrices.Unbind(program.SceneMatrices.Binding);
  sharedParams.Unbind(program.Uniforms[(int)UniformIndexHand::SharedParms].Binding);
  MaterialVertexBuffer.Unbind(program.Uniforms[(int)UniformIndexHand::VertMaterials].Binding);

  glBindVertexArray(0);
  glUseProgram(0);
}
