#include "LipSync.h"
#include <cassert>
#include "Utilities/Log.h"

using namespace std;

namespace {
const char* ConvertLipSyncResult(ovrLipSyncResult result) {
  switch (result) {
    case ovrLipSyncError_Unknown:
      return "Unknown Error";
    case ovrLipSyncError_CannotCreateContext:
      return "Cannot create context";
    case ovrLipSyncError_InvalidParam:
      return "Invalid Param";
    case ovrLipSyncError_BadSampleRate:
      return "Bad Sample Rate";
    case ovrLipSyncError_MissingDLL:
      return "Missing DLL";
    case ovrLipSyncError_BadVersion:
      return "Bad Version";
    case ovrLipSyncError_UndefinedFunction:
      return "Undefined Function";
    default:
      return "API Unknown Error";
      break;
  }
}
} // namespace

ovrLipSync::ovrLipSync() : Context(0) {
  auto result = ovrLipSync_Initialize(48000, 3840); // 80ms worth
  if (result != ovrLipSyncSuccess) {
    OVR_WARN("failed to initialize: %s", ConvertLipSyncResult(result));
    return;
  }

  result = ovrLipSync_CreateContextEx(
      &Context, ovrLipSyncContextProvider::ovrLipSyncContextProvider_Enhanced, 0, true);
  if (result != ovrLipSyncSuccess) {
    OVR_FAIL("failed to create context: %s", ConvertLipSyncResult(result));
  }

  result =
      ovrLipSync_SendSignal(Context, ovrLipSyncSignals::ovrLipSyncSignals_VisemeSmoothing, 0, 0);
  if (result != ovrLipSyncSuccess) {
    OVR_FAIL("failed to set viseme smoothing: %s", ConvertLipSyncResult(result));
  }

  CurrentVisemes(
      [](vector<float>& visemes) { fill_n(back_inserter(visemes), ovrLipSyncViseme_Count, 0.0f); });
  Initialized.store(true);
}

ovrLipSync::~ovrLipSync() {
  if (!Initialized.load()) {
    return;
  }
  auto result = ovrLipSync_DestroyContext(Context);
  if (result != ovrLipSyncSuccess) {
    OVR_WARN("failed to destroy context: %s", ConvertLipSyncResult(result));
  }

  result = ovrLipSync_Shutdown();
  if (result != ovrLipSyncSuccess) {
    OVR_WARN("failed to shutdown: %s", ConvertLipSyncResult(result));
  }
}

void ovrLipSync::ProcessAudio(const int16_t* buffer, uint32_t numSamples) {
  if (!Initialized.load()) {
    return;
  }
  if (CurrentlyProcessing.load()) {
    return;
  }
  CurrentlyProcessing.store(true);
  // OVRLipSync docs dont say if it takes a copy of the buffer or not, so ill just make a copy to be
  // safe.
  struct Capture {
    ovrLipSync* Self;
    vector<int16_t> Buffer;
  };
  Capture* capture = new Capture{};
  capture->Self = this;
  capture->Buffer.insert(begin(capture->Buffer), buffer, buffer + numSamples);
  auto result = ovrLipSync_ProcessFrameAsync(
      Context,
      capture->Buffer.data(),
      (int)capture->Buffer.size(),
      ovrLipSyncAudioDataType::ovrLipSyncAudioDataType_S16_Mono,
      [](void* capture, const ovrLipSyncFrame* pFrame, ovrLipSyncResult result) {
        reinterpret_cast<Capture*>(capture)->Self->OnFinishedProcessingAudio(pFrame, result);

        reinterpret_cast<Capture*>(capture)->Self->CurrentlyProcessing.store(false);
        delete reinterpret_cast<Capture*>(capture);
      },
      capture);
  if (result != ovrLipSyncSuccess) {
    OVR_WARN("ovrLipSync_ProcessFrameAsync failed: %s", ConvertLipSyncResult(result));
  }
}

void ovrLipSync::OnFinishedProcessingAudio(const ovrLipSyncFrame* pFrame, ovrLipSyncResult result) {
  if (result != ovrLipSyncSuccess) {
    OVR_WARN("ovrLipSync_ProcessFrameAsync failed on callback: %s", ConvertLipSyncResult(result));
    return;
  }
  CurrentVisemes([pFrame](vector<float>& visemes) {
    visemes.clear();
    visemes.reserve(pFrame->visemesLength);
    copy(pFrame->visemes, pFrame->visemes + pFrame->visemesLength, back_inserter(visemes));
  });
}

uint32_t ovrLipSync::GetVisemeCount() const {
  return ovrLipSyncViseme_Count;
}

std::vector<float> ovrLipSync::GetVisemes() const {
  std::vector<float> result;
  CurrentVisemes([&result](const vector<float>& visemes) {
    assert(visemes.size() <= ovrLipSyncViseme_Count);
    for (auto it = visemes.begin(); it != visemes.end(); ++it) {
      result.push_back(*it);
    }
  });
  return result;
}
