#include "Mesh.h"

#include "TypeConversions.h"
#include "Utilities/GlCore.h"
#include "Utilities/GlProgram.h"
#include "VertexBufferWriter.h"

#include <algorithm>
#include <cassert>
#include <type_traits>
#include <utility>

using namespace std;
using namespace ovrUtilities;
using namespace glm;

namespace {
constexpr ptrdiff_t PositionOffset = 0;
constexpr ptrdiff_t NormalOffset = PositionOffset + sizeof(float) * 3;
constexpr ptrdiff_t TangetOffset = NormalOffset + sizeof(uint32_t);
constexpr ptrdiff_t UVOffset = TangetOffset + sizeof(uint32_t);
constexpr ptrdiff_t JointIndicesOffset = UVOffset + sizeof(float) * 2;
constexpr ptrdiff_t JointWeightsOffset = JointIndicesOffset + sizeof(uint8_t) * 2;
constexpr ptrdiff_t RenderPartOffset = JointWeightsOffset + sizeof(uint8_t) * 2;
constexpr ptrdiff_t VertexStrideV1 =
    JointWeightsOffset + sizeof(uint8_t) * 2; // doesnt include render part
constexpr ptrdiff_t VertexStrideV2 = RenderPartOffset + 4;

static_assert(
    VertexStrideV1 % 4 == 0 && VertexStrideV2 % 4 == 0,
    "vertex size should be a multiple of 4 bytes, pad as neccessary");
} // namespace

template <typename T>
void ovrMesh::Init(const ovrAvatarAsset* asset, const T* mesh) {
  constexpr bool isV2 = is_same_v<decay_t<T>, ovrAvatarMeshAssetDataV2>;

  // TODO: Not currently rendering any submeshes, except for submesh 0, the primary mesh
  uint32_t smCount = ovrAvatarAsset_GetSubmeshCount(asset);
  assert(smCount >= 1);
  IndexCount = ovrAvatarAsset_GetSubmeshLastIndex(asset, 0);

  // Create Index buffer
  glGenBuffers(1, &IndexBuffer);

  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, IndexBuffer);
  glBufferData(
      GL_ELEMENT_ARRAY_BUFFER, IndexCount * sizeof(uint16_t), mesh->indexBuffer, GL_STATIC_DRAW);
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

  // Create Vertex Buffer, interleaved format
  // Going positions will be written as 3 floats
  // normals and tangents will be written in 10_10_10_2 format
  // Uv's will be written as 2 floats. Uv's are not limited to -1 to 1 range. so need full range
  // floats Bone weights are written as 8 bit normalized fixed point floats. Only need 2 bones. Bone
  // indices are written as 8 bit integers. Only need 2 bones.
  VertexBufferWriter vertexWriter;
  for (uint32_t i = 0; i < mesh->vertexCount; ++i) {
    vertexWriter.PackPosition(
        mesh->vertexBuffer[i].x, mesh->vertexBuffer[i].y, mesh->vertexBuffer[i].z);
    vertexWriter.PackFrenet(
        mesh->vertexBuffer[i].nx, mesh->vertexBuffer[i].ny, mesh->vertexBuffer[i].nz);
    vertexWriter.PackFrenet(
        mesh->vertexBuffer[i].tx, mesh->vertexBuffer[i].ty, mesh->vertexBuffer[i].tz);
    vertexWriter.PackUV(mesh->vertexBuffer[i].u, mesh->vertexBuffer[i].v);
    vertexWriter.PackJointIndices(
        mesh->vertexBuffer[i].blendIndices[0], mesh->vertexBuffer[i].blendIndices[1]);
    vertexWriter.PackJointWeights(
        mesh->vertexBuffer[i].blendWeights[0], mesh->vertexBuffer[i].blendWeights[1]);
    if constexpr (isV2) {
      vertexWriter.PackRenderPart(mesh->vertexBuffer[i].r);
    }
  }

  glGenBuffers(1, &VertexBuffer);
  glBindBuffer(GL_ARRAY_BUFFER, VertexBuffer);
  glBufferData(GL_ARRAY_BUFFER, size(vertexWriter), data(vertexWriter), GL_STATIC_DRAW);
  glBindBuffer(GL_ARRAY_BUFFER, 0);

  // create the vertex array
  glGenVertexArrays(1, &VertexArray);
  glBindVertexArray(VertexArray);
  glBindBuffer(GL_ARRAY_BUFFER, VertexBuffer);
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, IndexBuffer);

  uint32_t vertexStride = isV2 ? VertexStrideV2 : VertexStrideV1;

  glEnableVertexAttribArray(VERTEX_ATTRIBUTE_LOCATION_POSITION);
  glVertexAttribPointer(
      VERTEX_ATTRIBUTE_LOCATION_POSITION,
      3,
      GL_FLOAT,
      GL_FALSE,
      (GLsizei)vertexStride,
      (void*)PositionOffset);
  glEnableVertexAttribArray(VERTEX_ATTRIBUTE_LOCATION_NORMAL);
  glVertexAttribPointer(
      VERTEX_ATTRIBUTE_LOCATION_NORMAL,
      4,
      GL_INT_2_10_10_10_REV,
      GL_TRUE,
      (GLsizei)vertexStride,
      (void*)NormalOffset);
  glEnableVertexAttribArray(VERTEX_ATTRIBUTE_LOCATION_TANGENT);
  glVertexAttribPointer(
      VERTEX_ATTRIBUTE_LOCATION_TANGENT,
      4,
      GL_INT_2_10_10_10_REV,
      GL_TRUE,
      (GLsizei)vertexStride,
      (void*)TangetOffset);
  glEnableVertexAttribArray(VERTEX_ATTRIBUTE_LOCATION_UV0);
  glVertexAttribPointer(
      VERTEX_ATTRIBUTE_LOCATION_UV0, 2, GL_FLOAT, GL_TRUE, (GLsizei)vertexStride, (void*)UVOffset);
  glEnableVertexAttribArray(VERTEX_ATTRIBUTE_LOCATION_JOINT_INDICES);
  glVertexAttribIPointer(
      VERTEX_ATTRIBUTE_LOCATION_JOINT_INDICES,
      2,
      GL_UNSIGNED_BYTE,
      (GLsizei)vertexStride,
      (void*)JointIndicesOffset);
  glEnableVertexAttribArray(VERTEX_ATTRIBUTE_LOCATION_JOINT_WEIGHTS);
  glVertexAttribPointer(
      VERTEX_ATTRIBUTE_LOCATION_JOINT_WEIGHTS,
      2,
      GL_UNSIGNED_BYTE,
      GL_TRUE,
      (GLsizei)vertexStride,
      (void*)JointWeightsOffset);
  if constexpr (isV2) {
    glEnableVertexAttribArray(VERTEX_ATTRIBUTE_LOCATION_INDICES);
    glVertexAttribIPointer(
        VERTEX_ATTRIBUTE_LOCATION_INDICES,
        4,
        GL_UNSIGNED_BYTE,
        (GLsizei)vertexStride,
        (void*)RenderPartOffset);
  }
  glBindVertexArray(0);

  glDisableVertexAttribArray(VERTEX_ATTRIBUTE_LOCATION_POSITION);
  glDisableVertexAttribArray(VERTEX_ATTRIBUTE_LOCATION_NORMAL);
  glDisableVertexAttribArray(VERTEX_ATTRIBUTE_LOCATION_TANGENT);
  glDisableVertexAttribArray(VERTEX_ATTRIBUTE_LOCATION_UV0);
  glDisableVertexAttribArray(VERTEX_ATTRIBUTE_LOCATION_JOINT_INDICES);
  glDisableVertexAttribArray(VERTEX_ATTRIBUTE_LOCATION_JOINT_WEIGHTS);
  glDisableVertexAttribArray(VERTEX_ATTRIBUTE_LOCATION_INDICES);

  // now compute the inverse bind pose
  assert(mesh->skinnedBindPose.jointCount <= MAX_JOINTS);
  const uint32_t maxJoints = std::min<uint32_t>(mesh->skinnedBindPose.jointCount, MAX_JOINTS);
  InverseBindPose.resize(maxJoints);
  for (uint32_t i = 0; i < maxJoints; i++) {
    mat4 worldPose;
    mat4 local = Convert(mesh->skinnedBindPose.jointTransform[i]);
    const int parentIndex = mesh->skinnedBindPose.jointParents[i];
    if (parentIndex < 0) {
      worldPose = local;
    } else {
      worldPose = InverseBindPose[parentIndex] * local;
    }
    InverseBindPose[i] = worldPose;
  }
  for (auto& mat : InverseBindPose) {
    mat = glm::inverse(mat);
  }
}

ovrMesh::ovrMesh(const ovrAvatarAsset* asset, const ovrAvatarMeshAssetData* mesh) {
  Init(asset, mesh);
}

ovrMesh::ovrMesh(const ovrAvatarAsset* asset, const ovrAvatarMeshAssetDataV2* mesh) {
  Init(asset, mesh);
}

ovrMesh::ovrMesh(ovrMesh&& other) {
  *this = move(other);
}

ovrMesh& ovrMesh::operator=(ovrMesh&& other) {
  VertexArray = other.VertexArray;
  VertexBuffer = other.VertexBuffer;
  IndexBuffer = other.IndexBuffer;
  IndexCount = other.IndexCount;
  InverseBindPose = move(other.InverseBindPose);

  other.VertexArray = 0;
  other.VertexBuffer = 0;
  other.IndexBuffer = 0;
  other.IndexCount = 0;

  return *this;
}

ovrMesh::~ovrMesh() {
  if (VertexArray != 0) {
    glDeleteVertexArrays(1, &VertexArray);
    glDeleteBuffers(1, &IndexBuffer);
    glDeleteBuffers(1, &VertexBuffer);
  }
}
