#include "Shaders.h"

#include <iterator>

using namespace std;
using namespace ovrUtilities;

const ovrUtilities::ovrGlProgram& GetBodyProgram() {
  static const char* vertexShader = R"(
    precision highp float;
    precision highp samplerBuffer;
    precision highp sampler2D;

    in vec4 Position;
    in vec3 Normal;
    in vec3 Tangent;
    in vec2 TexCoord;
    in ivec2 JointIndices;
    in vec2 JointWeights;
    in ivec4 Indices;

    out vec4 WorldViewDir;
    out vec4 WorldLightDir;
    out vec4 VextexNormal;
    out vec3 VertexTangent;

    struct MatStruct
    {
      ivec4 BSConstants; //x has the size(in verts, i.e. 3 texels) of a single blendshape. y has number blend shapes used, always a multiple of 4
      // Currently 16 blend shapes to blend max. std140 layout, so cant
      // use an array of 16 ints/floats here :(
      ivec4 BSIndices[4]; //which blend shapes to blend
      vec4 BSWeights[4]; //blend shape weights
      mat4 Bones[9]; //Skinning bones
    };

    layout (std140) uniform VertMaterials
    {
      MatStruct Data;
    } mats;

    struct SharedStruct
    {
      vec3 ViewPosition;
      vec3 LightPosition;
    };

    //Parameters that are the same for every draw call, shared by multiple shaders
    layout (std140) uniform SharedParms
    {
      SharedStruct Data;
    } sharedp;

    //3 texels for each vert
    uniform samplerBuffer BlendShapes;
    
    const int BODY_COMPONENT = 0;

    void AddBlendShapes(in samplerBuffer blendShapes, in int size, in ivec4 indices, in vec4 weights,
                        inout vec4 position, inout vec3 normal, inout vec3 tangent ){
      int bshapeVert = indices.x * size + gl_VertexID;
      position.xyz += weights.x * texelFetch(blendShapes, 3*bshapeVert).xyz;
      normal.xyz += weights.x * texelFetch(blendShapes, 3*bshapeVert + 1).xyz;
      tangent.xyz += weights.x * texelFetch(blendShapes, 3*bshapeVert + 2).xyz;

      bshapeVert = indices.y * size + gl_VertexID;
      position.xyz += weights.y * texelFetch(blendShapes, 3*bshapeVert).xyz;
      normal.xyz += weights.y * texelFetch(blendShapes, 3*bshapeVert + 1).xyz;
      tangent.xyz += weights.y * texelFetch(blendShapes, 3*bshapeVert + 2).xyz;

      bshapeVert = indices.z * size + gl_VertexID;
      position.xyz += weights.z * texelFetch(blendShapes, 3*bshapeVert).xyz;
      normal.xyz += weights.z * texelFetch(blendShapes, 3*bshapeVert + 1).xyz;
      tangent.xyz += weights.z * texelFetch(blendShapes, 3*bshapeVert + 2).xyz;

      bshapeVert = indices.w * size + gl_VertexID;
      position.xyz += weights.w * texelFetch(blendShapes, 3*bshapeVert).xyz;
      normal.xyz += weights.w * texelFetch(blendShapes, 3*bshapeVert + 1).xyz;
      tangent.xyz += weights.w * texelFetch(blendShapes, 3*bshapeVert + 2).xyz;
    }

    void CalcPoseAnimated(in vec4 position, in vec3 normal, in vec3 tangent, in mat4 joint0,
                          in mat4 joint1, in vec2 weights,
                          out vec4 vertPose, out vec4 normPose, out vec4 tangentPose) {
      vertPose = joint0 * position * weights.x;
      vertPose += joint1 * position * weights.y;

      vec4 normal4 = vec4( normal, 0.0 );
      normPose = joint0 * normal4 * weights.x;
      normPose += joint1 * normal4 * weights.y;
      normPose = normalize( normPose );

      vec4 tangent4 = vec4( tangent, 0.0 );
      tangentPose = joint0 * tangent4 * weights.x;
      tangentPose += joint1 * tangent4 * weights.y;
      tangentPose = normalize( tangentPose );
    }

    void main()
    {
      vec4 position = Position;
      vec3 normal = Normal;
      vec3 tangent = Tangent;

      int componentIndex = Indices.x;
     
      //Handle blend shapes
      for(int i = 0;i <  mats.Data.BSConstants.y; ++i) {
        AddBlendShapes(BlendShapes, mats.Data.BSConstants.x, mats.Data.BSIndices[i],
                      mats.Data.BSWeights[i], position, normal, tangent);
      }

      mat4 joint0 = mats.Data.Bones[JointIndices.x];
      mat4 joint1 = mats.Data.Bones[JointIndices.y];

      vec4 vertexPose;
      vec4 normalPose;
      vec4 tangentPose;
      CalcPoseAnimated(position, normal, tangent, joint0, joint1, JointWeights, vertexPose, normalPose, tangentPose);

      vec3 vertexWorldPos;
      gl_Position = TransformVertex( vertexPose );
      vertexWorldPos = ( ModelMatrix * vertexPose ).xyz;
      normalPose = ModelMatrix * normalPose;
      tangentPose = ModelMatrix * tangentPose;

      VextexNormal.xyz = normalPose.xyz;
      VertexTangent.xyz = tangentPose.xyz;

      WorldViewDir.xyz = normalize( sharedp.Data.ViewPosition - vertexWorldPos.xyz );
      WorldLightDir.xyz = normalize( sharedp.Data.LightPosition - vertexWorldPos.xyz );

      WorldViewDir.w = TexCoord.x;
      WorldLightDir.w = TexCoord.y;
      VextexNormal.w = float(componentIndex);
    }
	)";

  static const char* fragmentShader = R"(
    precision mediump float;
    precision mediump sampler2D;
    precision mediump sampler2DArray;
    precision mediump samplerCube;

    in vec4 WorldViewDir;
    in vec4 WorldLightDir;
    in vec4 VextexNormal;
    in vec3 VertexTangent;

    //texture parameters
    uniform sampler2DArray Albedo;
    uniform sampler2DArray Normal;
    uniform sampler2DArray Metallic;
    uniform samplerCube EnvironmentMap;

    //From Avatar SDK - OVR_Avatar.h ovrAvatarBodyPartType::ovrAvatarBodyPartType_Count
    #define ovrAvatarBodyPartType_Count 5

    struct MatStruct
    {
      vec4 AlbedoTint[ovrAvatarBodyPartType_Count]; // One per renderpart.
      vec4 Intensity[ovrAvatarBodyPartType_Count]; //Diffuse in r, Reflection in g, brow in b, lip smoothness in a
      vec4 Colors[7]; // Expressive Colors
    };

    layout (std140) uniform FragMaterials
    {
      MatStruct Data;
    } mats;

    //values of roughness texture for selecting a color. Because of
    //mips/texture compression, the value wont be exact, round to closest.
    // Iris 255/255
    // Sclera 238/255
    // Lashes 221/255
    // Lips 204/255
    // Gums 187/255
    // Teeth 170/255
    // Brows 153/255
    const float MASK_MIN = 153.0/255.0;
    const float MASK_DIVISOR = 17.0/255.0;
    const float MASK_COUNT = 7.0;

    // Brightness modifier order: Brows, Teeth, Gums, Lips, Lashes, Sclera, Iris
    const float MASK_BRIGHTNESS_MODIFIERS[7] = float[7](1.0, 1.0, 1.0, 1.0, 1.0, 1.1, 2.0);

    const int BODY_COMPONENT = 0;
    const int BROWS_COLOR_INDEX = 0;
    const int TEETH_COLOR_INDEX = 1;
    const int GUMS_COLOR_INDEX = 2;
    const int LIPS_COLOR_INDEX = 3;
    const int LASHES_COLOR_INDEX = 4;
    const int SCLERA_COLOR_INDEX = 5;
    const int IRIS_COLOR_INDEX = 6;

    const float BROWS_LASHES_DIFFUSEINTENSITY = 0.75;
    const float LIP_SMOOTHNESS_MULTIPLIER = 0.5;
    const float MAX_SMOOTHNESS = 0.5;

    vec3 CalcLightingExpressive(vec3 albedoColor, highp vec3 lightDir, highp vec3 viewDir, highp vec3 normalDir, float diffuseIntensity, float lipGlossiness, float eyeGlint) {
      //constants
      const float LIP_GLINT_SPREAD = 150.0;
      const float EYE_GLINT_SPREAD = 600.0;

      //preliminaries
      vec3 halfv = normalize(lightDir + viewDir);
      float ndotl = max(dot(normalDir, lightDir),0.0);
      float ndoth = max(dot(normalDir, halfv), 0.0);

      // final
      vec3 result;
      result.rgb = albedoColor;
      result.rgb += diffuseIntensity * ndotl;
      
      lipGlossiness *= lipGlossiness > 0.0 ? pow(ndoth, LIP_GLINT_SPREAD) : 0.0;
      eyeGlint *= eyeGlint > 0.0 ? pow(ndoth, EYE_GLINT_SPREAD) : 0.0;
      result.rgb += lipGlossiness + eyeGlint;

      return result;
    }

    void main()
    {
      // Normalize vectors
      vec3 viewDir = normalize(WorldViewDir.xyz);
      vec3 lightDir = normalize(WorldLightDir.xyz);
      vec3 vertNormal = normalize(VextexNormal.xyz);
      vec3 vertTangent = normalize(VertexTangent.xyz);
      vec3 bitangent = cross(vertNormal, vertTangent);
      mat3 tangentToWorld = mat3(vertTangent,
                  bitangent,
                  vertNormal);
      vec2 uv = vec2(WorldViewDir.w, WorldLightDir.w);

      int matIndex = int(round(VextexNormal.w));

      vec4 albedoTex = texture(Albedo, vec3(uv,matIndex));
      
      // Normal
      vec3 normalDir = texture(Normal, vec3(uv,matIndex)).rgb;
      normalDir = 2.0 * normalDir - 1.0;
      normalDir = normalize(tangentToWorld * normalDir);
      
      // Surface properties
      vec3 albedoTint = mats.Data.AlbedoTint[matIndex].rgb;
      float diffuseIntensity = mats.Data.Intensity[matIndex].r;

      vec4 roughnessTex = texture(Metallic, vec3(uv,matIndex));
      float roughness = roughnessTex.a;
      
      // Mask colors
      int colorIndex = int(clamp(round((roughnessTex.b - MASK_MIN) / MASK_DIVISOR), 0.0, MASK_COUNT-1.0));
      vec3 colorMask = mats.Data.Colors[colorIndex].rgb;
      colorMask *= MASK_BRIGHTNESS_MODIFIERS[colorIndex];
      float mixValue = matIndex == BODY_COMPONENT ? roughnessTex.g : 0.0;
      mixValue *= mats.Data.Colors[colorIndex].a;
      albedoTint = mix(albedoTint, colorMask, mixValue);
        
      // Mask regions
      float maskBrowsScalar = colorIndex == BROWS_COLOR_INDEX ? mixValue : 0.0;
      float maskLashesScalar = colorIndex == LASHES_COLOR_INDEX ? mixValue : 0.0;
      float maskLipsScalar = colorIndex == LIPS_COLOR_INDEX ? mixValue : 0.0;
      float maskIrisScalar = colorIndex == IRIS_COLOR_INDEX ? mixValue : 0.0;
      float maskScleraScalar = colorIndex == SCLERA_COLOR_INDEX ? mixValue : 0.0;
      float maskTeethScalar = colorIndex == TEETH_COLOR_INDEX ? mixValue : 0.0;
      float maskGumsScalar = colorIndex == GUMS_COLOR_INDEX ? mixValue : 0.0;

      // Lerp diffuseIntensity with roughness map. This clamps the min diffuseIntensity to the
      // per-component property, and allows it to vary with the texture value up to a maximum.
      diffuseIntensity = mix(diffuseIntensity, MAX_SMOOTHNESS, roughness);
      // Brows and lashes, teeth and gums modify DiffuseIntensity
      diffuseIntensity *= (1.0 - ((maskBrowsScalar + maskLashesScalar) * BROWS_LASHES_DIFFUSEINTENSITY));
      diffuseIntensity *= (1.0 - (maskTeethScalar + maskGumsScalar));

      vec3 albedoColor = colorIndex == LIPS_COLOR_INDEX
        ? mix(albedoTex.rgb, colorMask, mixValue)
        : albedoTex.rgb * albedoTint;

      highp vec3 reflection = reflect(-viewDir, normalDir);
      vec3 reflectionColor = texture(EnvironmentMap, reflection).rgb;
      // Smoothness multiplier on lip region
      float lipGlossiness = maskLipsScalar * (mats.Data.Intensity[matIndex].a * LIP_SMOOTHNESS_MULTIPLIER);
      float eyeGlint = clamp(maskIrisScalar + maskScleraScalar, 0.0, 1.0);
      // Calculate lighting for expressive head
      fragColor.rgb = CalcLightingExpressive(albedoColor, lightDir, viewDir, normalDir, diffuseIntensity, lipGlossiness, eyeGlint);
      // Add reflection with cubemap if available
      fragColor.rgb += vec3(roughness) * mats.Data.Intensity[matIndex].g * reflectionColor;

      fragColor.a = albedoTex.a;
    }
  )";

  ovrProgramParm shaderParms[] = {
      {"VertMaterials", ovrProgramParmType::BUFFER_UNIFORM},
      {"FragMaterials", ovrProgramParmType::BUFFER_UNIFORM},
      {"SharedParms", ovrProgramParmType::BUFFER_UNIFORM},
      {"Albedo", ovrProgramParmType::TEXTURE_SAMPLED},
      {"Normal", ovrProgramParmType::TEXTURE_SAMPLED},
      {"Metallic", ovrProgramParmType::TEXTURE_SAMPLED},
      {"EnvironmentMap", ovrProgramParmType::TEXTURE_SAMPLED},
      {"BlendShapes", ovrProgramParmType::TEXTURE_SAMPLED},
  };

  static ovrGlProgram Program(
      "#extension GL_EXT_texture_buffer : require\n",
      vertexShader,
      "",
      fragmentShader,
      shaderParms,
      size(shaderParms));

  return Program;
}

const ovrUtilities::ovrGlProgram& GetHandProgram() {
  static const char* vertexShader = R"(
    precision highp float;
    precision highp sampler2D;

    in vec4 Position;
    in vec3 Normal;
    in vec3 Tangent;
    in vec2 TexCoord;
    in ivec2 JointIndices;
    in vec2 JointWeights;

    out vec4 WorldViewDir;
    out vec4 WorldLightDir;
    out vec3 VextexNormal;
    out vec3 VertexTangent;

    struct MatStruct
    {
      mat4 Bones[28]; //Skinning bones
    };

    layout (std140) uniform VertMaterials
    {
      MatStruct Data;
    } mats;

    struct SharedStruct
    {
      vec3 ViewPosition;
      vec3 LightPosition;
    };

    //Parameters that are the same for every draw call, shared by multiple shaders
    layout (std140) uniform SharedParms
    {
      SharedStruct Data;
    } sharedp;
    
    void CalcPoseAnimated(in vec4 position, in vec3 normal, in vec3 tangent, in mat4 joint0,
                          in mat4 joint1, in vec2 weights,
                          out vec4 vertPose, out vec4 normPose, out vec4 tangentPose) {
      vertPose = joint0 * position * weights.x;
      vertPose += joint1 * position * weights.y;

      vec4 normal4 = vec4( normal, 0.0 );
      normPose = joint0 * normal4 * weights.x;
      normPose += joint1 * normal4 * weights.y;
      normPose = normalize( normPose );

      vec4 tangent4 = vec4( tangent, 0.0 );
      tangentPose = joint0 * tangent4 * weights.x;
      tangentPose += joint1 * tangent4 * weights.y;
      tangentPose = normalize( tangentPose );
    }

    void main()
    {
      vec4 position = Position;
      vec3 normal = Normal;
      vec3 tangent = Tangent;
     
      mat4 joint0 = mats.Data.Bones[JointIndices.x];
      mat4 joint1 = mats.Data.Bones[JointIndices.y];

      vec4 vertexPose;
      vec4 normalPose;
      vec4 tangentPose;
      CalcPoseAnimated(position, normal, tangent, joint0, joint1, JointWeights, vertexPose, normalPose, tangentPose);

      vec3 vertexWorldPos;
      gl_Position = TransformVertex( vertexPose );
      vertexWorldPos = ( ModelMatrix * vertexPose ).xyz;
      normalPose = ModelMatrix * normalPose;
      tangentPose = ModelMatrix * tangentPose;

      VextexNormal.xyz = normalPose.xyz;
      VertexTangent.xyz = tangentPose.xyz;

      WorldViewDir.xyz = normalize( sharedp.Data.ViewPosition - vertexWorldPos.xyz );
      WorldLightDir.xyz = normalize( sharedp.Data.LightPosition - vertexWorldPos.xyz );

      WorldViewDir.w = TexCoord.x;
      WorldLightDir.w = TexCoord.y;
    }
	)";

  static const char* fragmentShader = R"(
    precision mediump float;
    precision mediump sampler2D;
    precision mediump sampler2DArray;
    precision mediump samplerCube;

    in vec4 WorldViewDir;
    in vec4 WorldLightDir;
    in vec3 VextexNormal;
    in vec3 VertexTangent;

    //texture parameters
    uniform sampler2DArray Albedo;
    uniform sampler2DArray Normal;

    uniform vec4 AlbedoTint;
    uniform int Hand; //0 = left, 1 = right

    vec3 CalcLighting(vec3 albedoColor, highp vec3 lightDir, highp vec3 viewDir, highp vec3 normalDir, float diffuseIntensity, float rimPower) {
      float ndotl = clamp(dot(normalDir, lightDir),0.0,1.0);

      //final
      vec3 result;
      result.rgb = albedoColor;
      result.rgb += diffuseIntensity * ndotl;

      return result;
    }

    void main()
    {
      // Normalize vectors
      vec3 viewDir = normalize(WorldViewDir.xyz);
      vec3 lightDir = normalize(WorldLightDir.xyz);
      vec3 vertNormal = normalize(VextexNormal.xyz);
      vec3 vertTangent = normalize(VertexTangent.xyz);
      vec3 bitangent = cross(vertNormal, vertTangent);
      mat3 tangentToWorld = mat3(vertTangent,
                  bitangent,
                  vertNormal);
      vec2 uv = vec2(WorldViewDir.w, WorldLightDir.w);

      vec4 albedoTex = texture(Albedo, vec3(uv, Hand));
      
      // Normal
      vec3 normalDir = texture(Normal, vec3(uv, Hand)).rgb;
      normalDir = 2.0 * normalDir - 1.0;
      normalDir = normalize(tangentToWorld * normalDir);
            
      vec3 albedoColor = albedoTex.rgb * AlbedoTint.rgb;

      fragColor.rgb = CalcLighting(albedoColor, lightDir, viewDir, normalDir, 0.3, 5.0);

      fragColor.a = albedoTex.a;
    }
  )";

  ovrProgramParm shaderParms[] = {
      {"VertMaterials", ovrProgramParmType::BUFFER_UNIFORM},
      {"SharedParms", ovrProgramParmType::BUFFER_UNIFORM},
      {"Albedo", ovrProgramParmType::TEXTURE_SAMPLED},
      {"Normal", ovrProgramParmType::TEXTURE_SAMPLED},
      {"AlbedoTint", ovrProgramParmType::FLOAT_VECTOR4},
      {"Hand", ovrProgramParmType::INT},
  };

  static ovrGlProgram Program("", vertexShader, "", fragmentShader, shaderParms, size(shaderParms));

  return Program;
}

const ovrUtilities::ovrGlProgram& GetControllerProgramV1() {
  static const char* vertexShader = R"(
    precision highp float;
    precision highp sampler2D;

    in vec4 Position;
    in vec3 Normal;
    in vec2 TexCoord;
    in ivec2 JointIndices;
    in vec2 JointWeights;

    out vec4 WorldViewDir;
    out vec4 WorldLightDir;
    out vec3 VextexNormal;

    struct MatStruct
    {
      mat4 Bones[8]; //Skinning bones
    };

    layout (std140) uniform VertMaterials
    {
      MatStruct Data;
    } mats;

    struct SharedStruct
    {
      vec3 ViewPosition;
      vec3 LightPosition;
    };

    //Parameters that are the same for every draw call, shared by multiple shaders
    layout (std140) uniform SharedParms
    {
      SharedStruct Data;
    } sharedp;
    
    void CalcPoseAnimated(in vec4 position, in vec3 normal,in mat4 joint0,
                          in mat4 joint1, in vec2 weights,
                          out vec4 vertPose, out vec4 normPose) {
      vertPose = joint0 * position * weights.x;
      vertPose += joint1 * position * weights.y;

      vec4 normal4 = vec4( normal, 0.0 );
      normPose = joint0 * normal4 * weights.x;
      normPose += joint1 * normal4 * weights.y;
      normPose = normalize( normPose );
    }

    void main()
    {
      vec4 position = Position;
      vec3 normal = Normal;
     
      mat4 joint0 = mats.Data.Bones[JointIndices.x];
      mat4 joint1 = mats.Data.Bones[JointIndices.y];

      vec4 vertexPose;
      vec4 normalPose;
      CalcPoseAnimated(position, normal, joint0, joint1, JointWeights, vertexPose, normalPose);

      vec3 vertexWorldPos;
      gl_Position = TransformVertex( vertexPose );
      vertexWorldPos = ( ModelMatrix * vertexPose ).xyz;
      normalPose = ModelMatrix * normalPose;

      VextexNormal.xyz = normalPose.xyz;

      WorldViewDir.xyz = normalize( sharedp.Data.ViewPosition - vertexWorldPos.xyz );
      WorldLightDir.xyz = normalize( sharedp.Data.LightPosition - vertexWorldPos.xyz );

      WorldViewDir.w = TexCoord.x;
      WorldLightDir.w = TexCoord.y;
    }
	)";

  static const char* fragmentShader = R"(
    precision mediump float;
    precision mediump sampler2D;
    precision mediump sampler2DArray;
    precision mediump samplerCube;

    in vec4 WorldViewDir;
    in vec4 WorldLightDir;
    in vec3 VextexNormal;

    //texture parameters
    uniform sampler2DArray Albedo;
    uniform sampler2DArray Metallic;
    uniform samplerCube EnvironmentMap;

    uniform int Hand; //0 = left, 1 = right

    float G1V(float dotp, float k) {
      return 1.0/(dotp*(1.0-k)+k);
    }

    vec3 CalcLighting(vec3 albedoColor, vec4 metallicTex, highp vec3 lightDir, highp vec3 viewDir, highp vec3 normalDir, vec3 reflectionColor) {
      //constants
      const float ambientFactor = 0.2;
      float roughness = 1.0 - metallicTex.a;
      float metallic = metallicTex.r;
      float diffuseFactor = 0.98*(roughness);
      float specularFactor = 1.0 - diffuseFactor - ambientFactor;

      //preliminaries
      vec3 halfv = normalize(lightDir + viewDir);
      float ndotl = max(dot(lightDir, normalDir),0.0);
      float ndoth = max(dot(halfv, normalDir),0.0);
      float ldoth = max(dot(halfv, lightDir),0.0);
      float ndotv = max(dot(normalDir, viewDir),0.0);

      //ambient
      vec3 ambientLight = albedoColor*ambientFactor;

      //diffuse
      vec3 diffuseLight = diffuseFactor * ndotl*albedoColor;

      //specular
      float roughsq = roughness*roughness;

      //D
      float roughtp4 = roughsq*roughsq;
      float denom = ndoth * ndoth * (roughtp4-1.0) + 1.0f;
      float D = roughtp4/(3.1459*denom*denom);

      //F
      vec3 F0 = mix(vec3(0.2),albedoColor,metallic);
      vec3 fresnel = F0 + (1.0 - F0 )*pow(1.0-ldoth,5.0);

      //V
      float k = roughsq/2.0f;
      float V = G1V(ndotl,k)*G1V(ndotv,k);

      vec3 specularLight = fresnel * D * V * ndotl;

      //final
      return ambientLight + diffuseLight + specularLight + (reflectionColor*specularFactor*0.25);
    }

    void main()
    {
      // Normalize vectors
      vec3 viewDir = normalize(WorldViewDir.xyz);
      vec3 lightDir = normalize(WorldLightDir.xyz);
      vec3 normalDir = normalize(VextexNormal.xyz);
      vec2 uv = vec2(WorldViewDir.w, WorldLightDir.w);

      vec3 albedoTex = texture(Albedo, vec3(uv, Hand)).rgb;
      vec4 metallicTex = texture(Metallic, vec3(uv, Hand));

      vec3 reflection = reflect(-viewDir, normalDir);
      vec3 reflectionColor = texture(EnvironmentMap, reflection).rgb;
                  
      fragColor.rgb = CalcLighting(albedoTex, metallicTex, lightDir, viewDir, normalDir, reflectionColor);

      fragColor.a = 1.0;
    }
  )";

  ovrProgramParm shaderParms[] = {
      {"VertMaterials", ovrProgramParmType::BUFFER_UNIFORM},
      {"SharedParms", ovrProgramParmType::BUFFER_UNIFORM},
      {"Albedo", ovrProgramParmType::TEXTURE_SAMPLED},
      {"Metallic", ovrProgramParmType::TEXTURE_SAMPLED},
      {"EnvironmentMap", ovrProgramParmType::TEXTURE_SAMPLED},
      {"Hand", ovrProgramParmType::INT},
  };

  static ovrGlProgram Program("", vertexShader, "", fragmentShader, shaderParms, size(shaderParms));

  return Program;
}

const ovrUtilities::ovrGlProgram& GetControllerProgramV2() {
  static const char* vertexShader = R"(
    precision highp float;
    precision highp sampler2D;

    in vec4 Position;
    in vec3 Normal;
    in vec3 Tangent;
    in vec2 TexCoord;
    in ivec2 JointIndices;
    in vec2 JointWeights;

    out vec4 WorldViewDir;
    out vec4 WorldLightDir;
    out vec3 VextexNormal;
    out vec3 VertexTangent;

    struct MatStruct
    {
      mat4 Bones[8]; //Skinning bones
    };

    layout (std140) uniform VertMaterials
    {
      MatStruct Data;
    } mats;

    struct SharedStruct
    {
      vec3 ViewPosition;
      vec3 LightPosition;
    };

    //Parameters that are the same for every draw call, shared by multiple shaders
    layout (std140) uniform SharedParms
    {
      SharedStruct Data;
    } sharedp;
    
    void CalcPoseAnimated(in vec4 position, in vec3 normal, in vec3 tangent, in mat4 joint0,
                          in mat4 joint1, in vec2 weights,
                          out vec4 vertPose, out vec4 normPose, out vec4 tangentPose) {
      vertPose = joint0 * position * weights.x;
      vertPose += joint1 * position * weights.y;

      vec4 normal4 = vec4( normal, 0.0 );
      normPose = joint0 * normal4 * weights.x;
      normPose += joint1 * normal4 * weights.y;
      normPose = normalize( normPose );

      vec4 tangent4 = vec4( tangent, 0.0 );
      tangentPose = joint0 * tangent4 * weights.x;
      tangentPose += joint1 * tangent4 * weights.y;
      tangentPose = normalize( tangentPose );
    }

    void main()
    {
      vec4 position = Position;
      vec3 normal = Normal;
      vec3 tangent = Tangent;
     
      mat4 joint0 = mats.Data.Bones[JointIndices.x];
      mat4 joint1 = mats.Data.Bones[JointIndices.y];

      vec4 vertexPose;
      vec4 normalPose;
      vec4 tangentPose;
      CalcPoseAnimated(position, normal, tangent, joint0, joint1, JointWeights, vertexPose, normalPose, tangentPose);

      vec3 vertexWorldPos;
      gl_Position = TransformVertex( vertexPose );
      vertexWorldPos = ( ModelMatrix * vertexPose ).xyz;
      normalPose = ModelMatrix * normalPose;
      tangentPose = ModelMatrix * tangentPose;

      VextexNormal.xyz = normalPose.xyz;
      VertexTangent.xyz = tangentPose.xyz;

      WorldViewDir.xyz = normalize( sharedp.Data.ViewPosition - vertexWorldPos.xyz );
      WorldLightDir.xyz = normalize( sharedp.Data.LightPosition - vertexWorldPos.xyz );

      WorldViewDir.w = TexCoord.x;
      WorldLightDir.w = TexCoord.y;
    }
	)";

  static const char* fragmentShader = R"(
    precision mediump float;
    precision mediump sampler2D;
    precision mediump sampler2DArray;
    precision mediump samplerCube;

    in vec4 WorldViewDir;
    in vec4 WorldLightDir;
    in vec3 VextexNormal;
    in vec3 VertexTangent;

    //texture parameters
    uniform sampler2DArray Albedo;
    uniform sampler2DArray Normal;
    uniform sampler2DArray Metallic;
    uniform samplerCube EnvironmentMap;

    uniform int Hand; //0 = left, 1 = right

    float G1V(float dotp, float k) {
      return 1.0f/(dotp*(1.0f-k)+k);
    }

    vec3 CalcLighting(vec3 albedoColor, vec4 metallicTex, highp vec3 lightDir, highp vec3 viewDir, highp vec3 normalDir, vec3 reflectionColor) {
      //constants
      const float ambientFactor = 0.2;
      float roughness = 1.0 - metallicTex.a;
      float metallic = metallicTex.r;
      float diffuseFactor = 0.98*(roughness);
      float specularFactor = 1.0 - diffuseFactor - ambientFactor;

      //preliminaries
      vec3 halfv = normalize(lightDir + viewDir);
      float ndotl = max(dot(lightDir, normalDir),0.0);
      float ndoth = max(dot(halfv, normalDir),0.0);
      float ldoth = max(dot(halfv, lightDir),0.0);
      float ndotv = max(dot(normalDir, viewDir),0.0);

      //ambient
      vec3 ambientLight = albedoColor*ambientFactor;

      //diffuse
      vec3 diffuseLight = diffuseFactor * ndotl*albedoColor;

      //specular
      float roughsq = roughness*roughness;

      //D
      float roughtp4 = roughsq*roughsq;
      float denom = ndoth * ndoth * (roughtp4-1.0) + 1.0f;
      float D = roughtp4/(3.1459*denom*denom);

      //F
      vec3 F0 = mix(vec3(0.2),albedoColor,metallic);
      vec3 fresnel = F0 + (1.0 - F0 )*pow(1.0-ldoth,5.0);

      //V
      float k = roughsq/2.0f;
      float V = G1V(ndotl,k)*G1V(ndotv,k);

      vec3 specularLight = fresnel * D * V * ndotl;

      //final
      return ambientLight + diffuseLight + specularLight + (reflectionColor*specularFactor*0.25);
    }

    void main()
    {
      // Normalize vectors
      vec3 viewDir = normalize(WorldViewDir.xyz);
      vec3 lightDir = normalize(WorldLightDir.xyz);
      vec3 vertNormal = normalize(VextexNormal.xyz);
      vec3 vertTangent = normalize(VertexTangent.xyz);
      vec3 bitangent = cross(vertNormal, vertTangent);
      mat3 tangentToWorld = mat3(vertTangent,
                  bitangent,
                  vertNormal);
      vec2 uv = vec2(WorldViewDir.w, WorldLightDir.w);

      vec3 albedoTex = texture(Albedo, vec3(uv, Hand)).rgb;
      vec4 metallicTex = texture(Metallic, vec3(uv, Hand));
      
      // Normal
      vec3 normalDir = texture(Normal, vec3(uv, Hand)).rgb;
      normalDir = 2.0 * normalDir - 1.0;
      normalDir = normalize(tangentToWorld * normalDir);
            

      vec3 reflection = reflect(-viewDir, normalDir);
      vec3 reflectionColor = texture(EnvironmentMap, reflection).rgb;

      fragColor.rgb = CalcLighting(albedoTex, metallicTex, lightDir, viewDir, normalDir, reflectionColor);

      fragColor.a = 1.0;
    }
  )";

  ovrProgramParm shaderParms[] = {
      {"VertMaterials", ovrProgramParmType::BUFFER_UNIFORM},
      {"SharedParms", ovrProgramParmType::BUFFER_UNIFORM},
      {"Albedo", ovrProgramParmType::TEXTURE_SAMPLED},
      {"Metallic", ovrProgramParmType::TEXTURE_SAMPLED},
      {"EnvironmentMap", ovrProgramParmType::TEXTURE_SAMPLED},
      {"Hand", ovrProgramParmType::INT},
      {"Normal", ovrProgramParmType::TEXTURE_SAMPLED},
  };

  static ovrGlProgram Program("", vertexShader, "", fragmentShader, shaderParms, size(shaderParms));

  return Program;
}
