#include "TextureLoader.h"

#include "Utilities/Log.h"

#include <algorithm>

using namespace ovrUtilities;
using namespace std;

void LoadTexture(
    ovrGlTexture& texDest,
    uint32_t index,
    const ovrAvatarTextureAssetData* tex,
    bool srgb) {
  uint32_t format = 0;
  if (tex->format == ovrAvatarTextureFormat_ASTC_RGB_6x6_MIPMAPS) {
    format = srgb ? GL_COMPRESSED_SRGB8_ALPHA8_ASTC_6x6_KHR : GL_COMPRESSED_RGBA_ASTC_6x6_KHR;
  } else if (tex->format == ovrAvatarTextureFormat_DXT1) {
    format = srgb ? GL_COMPRESSED_SRGB_S3TC_DXT1_EXT : GL_COMPRESSED_RGB_S3TC_DXT1_EXT;
  } else if (tex->format == ovrAvatarTextureFormat_DXT5) {
    format = srgb ? GL_COMPRESSED_SRGB_ALPHA_S3TC_DXT5_EXT : GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
  } else {
    OVR_FAIL("unsupported texture format from Avatar SDK %d", tex->format);
  }

  // Textures are all the same size for a given platform. If this is first one we have seen, create
  // the texture array with the size of that texture. For combined meshes, we will render the avatar
  // in a single draw call, but we still have up to 5 parts, each with a seperate texture. So we
  // will store the textures in a texture array, one entry per part(body, clothing, hair, eyewear,
  // beard). For hands we will put 2 textures, one for left, one for right
  if (!texDest.IsValid()) {
    texDest = ovrGlTexture(GL_TEXTURE_2D_ARRAY, format, tex->sizeX, tex->sizeY, 5, tex->mipCount);
  }

  texDest.Bind([&](auto& texId) {
    uint32_t width = tex->sizeX;
    uint32_t height = tex->sizeY;
    const uint8_t* data = tex->textureData;

    for (unsigned int mip = 0; mip < tex->mipCount; mip++) {
      int32_t mipSize = 0;
      if (tex->format == ovrAvatarTextureFormat_ASTC_RGB_6x6_MIPMAPS) {
        int32_t blocksWide = (width + 5) / 6;
        int32_t blocksHigh = (height + 5) / 6;
        mipSize = 16 * blocksWide * blocksHigh;
      } else if (tex->format == ovrAvatarTextureFormat_DXT1) {
        int32_t blocksWide = (width + 3) / 4;
        int32_t blocksHigh = (height + 3) / 4;
        mipSize = 8 * blocksWide * blocksHigh;
      } else if (tex->format == ovrAvatarTextureFormat_DXT5) {
        int32_t blocksWide = (width + 3) / 4;
        int32_t blocksHigh = (height + 3) / 4;
        mipSize = 16 * blocksWide * blocksHigh;
      } else {
        OVR_FAIL("Invalid texture format");
      }

      glCompressedTexSubImage3D(
          GL_TEXTURE_2D_ARRAY, mip, 0, 0, index, width, height, 1, format, mipSize, data);

      data += mipSize;

      width /= 2;
      height /= 2;
      width = std::max(width, 1u);
      height = std::max(height, 1u);
    }

    glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_T, GL_REPEAT);

    glTexParameteri(
        GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MAX_ANISOTROPY_EXT, std::min(4, GetMaxAnisotropic()));
  });
}
