#include "GlBuffer.h"

#include <cassert>
#include <utility>
#include "GlCore.h"
#include "Log.h"

using namespace std;
using namespace ovrUtilities;

ovrGlBuffer::ovrGlBuffer() : Target(0), Buffer(0), Size(0) {}

ovrGlBuffer::~ovrGlBuffer() {
  Destroy();
}

ovrGlBuffer::ovrGlBuffer(const ovrGlBufferType_t type, const size_t dataSize) {
  switch (type) {
    case GLBUFFER_TYPE_UNIFORM:
      Target = GL_UNIFORM_BUFFER;
      break;
    case GLBUFFER_TYPE_TEXTURE_BUFFER:
      Target = GL_TEXTURE_BUFFER_EXT;
      break;
    case GLBUFFER_TYPE_SHADER_STORAGE:
      Target = GL_SHADER_STORAGE_BUFFER;
      break;
    case GLBUFFER_TYPE_ARRAY:
      Target = GL_ARRAY_BUFFER;
      break;
    default:
      OVR_WARN("unknown buffer type");
      Target = 0;
      break;
  }
  Size = dataSize;

  glGenBuffers(1, &Buffer);
  Bind();
  glBufferData(Target, Size, nullptr, GL_STATIC_DRAW);
  Unbind();
}

ovrGlBuffer::ovrGlBuffer(ovrGlBuffer&& other) {
  *this = move(other);
}

ovrGlBuffer& ovrGlBuffer::operator=(ovrGlBuffer&& other) {
  Destroy();
  Target = other.Target;
  Size = other.Size;
  Buffer = other.Buffer;

  other.Target = 0;
  other.Size = 0;
  other.Buffer = 0;

  return *this;
}

void ovrGlBuffer::Destroy() {
  if (Buffer != 0) {
    glDeleteBuffers(1, &Buffer);
    Buffer = 0;
  }
}

void ovrGlBuffer::Bind() const {
  glBindBuffer(Target, Buffer);
}

void ovrGlBuffer::Unbind() const {
  glBindBuffer(Target, 0);
}

void ovrGlBuffer::Bind(uint32_t index) const {
  glBindBufferBase(Target, index, Buffer);
}

void ovrGlBuffer::BindRange(uint32_t index, uint32_t offset, uint32_t size) const {
  glBindBufferRange(Target, index, Buffer, offset, size);
}

void ovrGlBuffer::Unbind(uint32_t index) const {
  glBindBufferBase(Target, index, 0);
}

void ovrGlBuffer::BindAsSSBO(uint32_t index) const {
  glBindBufferBase(GL_SHADER_STORAGE_BUFFER, index, Buffer);
}

void ovrGlBuffer::UnbindAsSSBO(uint32_t index) const {
  glBindBufferBase(GL_SHADER_STORAGE_BUFFER, index, 0);
}

void* ovrGlBuffer::Map() const {
  assert(Buffer != 0);

  void* data = NULL;
  Bind();
  data = glMapBufferRange(Target, 0, Size, GL_MAP_WRITE_BIT | GL_MAP_INVALIDATE_BUFFER_BIT);
  Unbind();

  if (data == NULL) {
    OVR_FAIL(
        "GlBuffer::MapBuffer: Failed to map buffer %d target %d size %zu", Buffer, Target, Size);
  }

  return data;
}

void ovrGlBuffer::Unmap() const {
  assert(Buffer != 0);

  Bind();
  if (!glUnmapBuffer(Target)) {
    OVR_WARN("GlBuffer::UnmapBuffer: Failed to unmap buffer.");
  }
  Unbind();
}
