#include "GlCore.h"
#include "Log.h"

#include <cstring>

using namespace ovrUtilities;

namespace {

#ifdef ANDROID
typedef void(
    GL_APIENTRYP PFNGLTEXBUFFEREXTPROC)(GLenum target, GLenum internalformat, GLuint buffer);
#endif

struct ovrExtensions {
  bool Initialized{false};
  bool TextureBuffers{false};
  int32_t MaxAnisotropic{1};

  PFNGLTEXBUFFEREXTPROC GlTexBufferEXT;
};

bool GL_ExtensionStringPresent(const char* extension, const char* allExtensions) {
  if (strstr(allExtensions, extension)) {
    OVR_LOG("Found: %s", extension);
    return true;
  } else {
    OVR_LOG("Not found: %s", extension);
    return false;
  }
}

const ovrExtensions& GetExtensions() {
  static ovrExtensions extensions;

  if (!extensions.Initialized) {
#ifdef ANDROID
    const char* extensionString = (const char*)glGetString(GL_EXTENSIONS);

    if (GL_ExtensionStringPresent("GL_EXT_texture_buffer", extensionString)) {
      extensions.TextureBuffers = true;
      extensions.GlTexBufferEXT = (PFNGLTEXBUFFEREXTPROC)eglGetProcAddress("glTexBufferEXT");

      GLint MaxTextureBufferSize = 0;
      glGetIntegerv(GL_MAX_TEXTURE_BUFFER_SIZE_EXT, &MaxTextureBufferSize);
      OVR_LOG("GL_MAX_TEXTURE_BUFFER_SIZE_EXT = %dK texels", (MaxTextureBufferSize / 1024));
    }

    if (GL_ExtensionStringPresent("GL_EXT_texture_filter_anisotropic", extensionString)) {
      glGetIntegerv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &extensions.MaxAnisotropic);
      OVR_LOG("MAX_TEXTURE_MAX_ANISOTROPY_EXT = %d", extensions.MaxAnisotropic);
    }
#else
    extensions.TextureBuffers = true;
    glGetIntegerv(GL_MAX_TEXTURE_MAX_ANISOTROPY, &extensions.MaxAnisotropic);
    OVR_LOG("MAX_TEXTURE_MAX_ANISOTROPY_EXT = %d", extensions.MaxAnisotropic);
#endif

    int32_t uniformBufAlign = 0;
    glGetIntegerv(GL_UNIFORM_BUFFER_OFFSET_ALIGNMENT, &uniformBufAlign);
    OVR_LOG("GL_UNIFORM_BUFFER_OFFSET_ALIGNMENT = %d", uniformBufAlign);

    extensions.Initialized = true;
  }
  return extensions;
}
} // namespace

bool ovrUtilities::HasTextureBufferExtension() {
  return GetExtensions().TextureBuffers;
}

void ovrUtilities::glTexBufferEXT(GLenum target, GLenum internalformat, GLuint buffer) {
  if (!HasTextureBufferExtension()) {
    OVR_FAIL("tried to call glTexBufferEXT on a platform that doesnt support it");
  }
#ifdef ANDROID
  GetExtensions().GlTexBufferEXT(target, internalformat, buffer);
#else
  glTexBuffer(target, internalformat, buffer);
#endif
}

int32_t ovrUtilities::GetMaxAnisotropic() {
  return GetExtensions().MaxAnisotropic;
}

void ovrUtilities::SetDebugCallback(
    void (*callback)(
        GLenum source,
        GLenum type,
        GLuint id,
        GLenum severity,
        GLsizei length,
        const GLchar* message,
        const void* usrPtr),
    const void* usrPtr) {
#ifdef ANDROID
  if (GL_ExtensionStringPresent("GL_KHR_debug", (const char*)glGetString(GL_EXTENSIONS))) {
    PFNGLDEBUGMESSAGECALLBACKKHRPROC glDebugMessageCallbackKHR_ =
        (PFNGLDEBUGMESSAGECALLBACKKHRPROC)eglGetProcAddress("glDebugMessageCallbackKHR");
    PFNGLDEBUGMESSAGECONTROLKHRPROC glDebugMessageControlKHR_ =
        (PFNGLDEBUGMESSAGECONTROLKHRPROC)eglGetProcAddress("glDebugMessageControlKHR");

    // set debug handler and enable output
    if (glDebugMessageCallbackKHR_ != nullptr) {
      glDebugMessageCallbackKHR_(callback, usrPtr);
      glEnable(GL_DEBUG_OUTPUT_KHR);
    }
    // attempt to enable all debug output
    if (glDebugMessageControlKHR_ != nullptr) {
      glDebugMessageControlKHR_(GL_DONT_CARE, GL_DONT_CARE, GL_DONT_CARE, 0, nullptr, GL_TRUE);
    }
  }
#else
  glDebugMessageCallback(callback, usrPtr);
  glEnable(GL_DEBUG_OUTPUT);
  glDebugMessageControl(GL_DONT_CARE, GL_DONT_CARE, GL_DONT_CARE, 0, nullptr, GL_TRUE);
#endif
}
