#include "GlProgram.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cassert>
#include <string>
#include "Format.h"
#include "GlCore.h"
#include "Log.h"

using namespace std;
using namespace ovrUtilities;

namespace {
bool UseMultiview = false;

// All GlPrograms implicitly get the VertexHeader
static const char* VertexHeader =
#if defined(ANDROID)
    R"=====(
#ifndef DISABLE_MULTIVIEW
 #define DISABLE_MULTIVIEW 0
#endif
#define NUM_VIEWS 2

#if defined( GL_OVR_multiview2 ) && ! DISABLE_MULTIVIEW
  #extension GL_OVR_multiview2 : require
  layout(num_views=NUM_VIEWS) in;
  #define VIEW_ID gl_ViewID_OVR
#else
  uniform lowp int ViewID;
  #define VIEW_ID ViewID
#endif

uniform highp mat4 ModelMatrix;
// Use a ubo to workaround corruption issue on Adreno 420+v300
// when uniform array of matrices used.
uniform SceneMatrices
{
	highp mat4 ViewMatrix[NUM_VIEWS];
	highp mat4 ProjectionMatrix[NUM_VIEWS];
} sm;
// Use a function macro for TransformVertex to workaround an issue on exynos8890+Android-M driver:
// gl_ViewID_OVR outside of the vertex main block causes VIEW_ID to be 0 for every view.
// NOTE: Driver fix has landed with Android-N.
//highp vec4 TransformVertex( highp vec4 oPos )
//{
//	highp vec4 hPos = sm.ProjectionMatrix[VIEW_ID] * ( sm.ViewMatrix[VIEW_ID] * ( ModelMatrix * oPos ) );
//	return hPos;
//}
#define TransformVertex(localPos) (sm.ProjectionMatrix[VIEW_ID] * ( sm.ViewMatrix[VIEW_ID] * ( ModelMatrix * localPos )))
)====="
#else
    R"=====(
    #define NUM_VIEWS 2
    uniform int ViewID;
    #define VIEW_ID ViewID
    uniform highp mat4 ModelMatrix;
    uniform SceneMatrices
    {
      highp mat4 ViewMatrix[NUM_VIEWS];
      highp mat4 ProjectionMatrix[NUM_VIEWS];
    } sm;
    highp vec4 TransformVertex( highp vec4 oPos )
    {
      highp vec4 hPos = sm.ProjectionMatrix[ VIEW_ID ] * ( sm.ViewMatrix[ VIEW_ID ] * ( ModelMatrix * oPos ) );
      return hPos;
    }
)====="
#endif
    ;

// All GlPrograms implicitly get the FragmentHeader
static const char* FragmentHeader =
#if defined(ANDROID)
    R"=====(
	  out mediump vec4 fragColor;
)====="
#else
    R"=====(
    out vec4 fragColor;
)====="
#endif
    ;

const char* GetShaderTypeString(GLenum shaderType) {
  static const char* vertex = "vertex";
  static const char* fragment = "fragment";
  static const char* compute = "compute";
  static const char* unknown = "unknown";

  switch (shaderType) {
    case GL_VERTEX_SHADER:
      return vertex;
    case GL_FRAGMENT_SHADER:
      return fragment;
    case GL_COMPUTE_SHADER:
      return compute;
    default:
      // Shouldn't ever happen
      assert(false);
      return unknown;
  }
}

GLuint CompileShader(GLenum shaderType, const char* directives, const char* src) {
  string srcString;

#if defined(ANDROID)
  // Valid versions for GL ES:
  // #version 100      -- ES 2.0
  // #version 300 es	 -- ES 3.0
  // #version 310 es	 -- ES 3.1
  srcString = "#version 310 es\n";
#else
  srcString = "#version 440\n";
#endif

  if (directives != nullptr) {
    srcString += directives;
  }

  if (shaderType != GL_COMPUTE_SHADER) {
    srcString += Format("#define DISABLE_MULTIVIEW %d\n", (UseMultiview) ? 0 : 1);
  }

  if (shaderType == GL_VERTEX_SHADER) {
    srcString += VertexHeader;
  } else if (shaderType == GL_FRAGMENT_SHADER) {
    srcString += FragmentHeader;
  }

  srcString += src;

  src = srcString.c_str();

  GLuint shader = glCreateShader(shaderType);

  const int numSources = 1;
  const char* srcs[1];
  srcs[0] = src;

  glShaderSource(shader, numSources, srcs, 0);
  glCompileShader(shader);

  GLint r;
  glGetShaderiv(shader, GL_COMPILE_STATUS, &r);
  if (r == GL_FALSE) {
    OVR_WARN("Compiling %s shader: ****** failed ******\n", GetShaderTypeString(shaderType));
    GLchar msg[1024];
    const char* sp = src;
    int charCount = 0;
    int line = 0;
    do {
      if (*sp != '\n') {
        msg[charCount++] = *sp;
        msg[charCount] = 0;
      }
      if (*sp == 0 || *sp == '\n' || charCount == 1023) {
        charCount = 0;
        line++;
        OVR_WARN("%03d  %s", line, msg);
        msg[0] = 0;
        if (*sp != '\n') {
          line--;
        }
      }
      sp++;
    } while (*sp != 0);
    if (charCount != 0) {
      line++;
      OVR_WARN("%03d  %s", line, msg);
    }
    glGetShaderInfoLog(shader, sizeof(msg), 0, msg);
    OVR_WARN("%s\n", msg);
    glDeleteShader(shader);
    return 0;
  }
  return shader;
}

} // namespace

ovrGlProgram::ovrGlProgram(
    const char* vertexDirectives,
    const char* vertexSrc,
    const char* fragmentDirectives,
    const char* fragmentSrc,
    const ovrProgramParm* parms,
    const int numParms) {
  //--------------------------
  // Compile and Create the Program
  //--------------------------

  assert(numParms <= 12);

  VertexShader = CompileShader(GL_VERTEX_SHADER, vertexDirectives, vertexSrc);
  if (VertexShader == 0) {
    Free();
    OVR_FAIL("Failed to compile vertex shader");
    return;
  }

  FragmentShader = CompileShader(GL_FRAGMENT_SHADER, fragmentDirectives, fragmentSrc);
  if (FragmentShader == 0) {
    Free();
    OVR_FAIL("Failed to compile fragment shader");
    return;
  }

  Program = glCreateProgram();
  glAttachShader(Program, VertexShader);
  glAttachShader(Program, FragmentShader);

  //--------------------------
  // Set attributes before linking
  //--------------------------

  glBindAttribLocation(Program, VERTEX_ATTRIBUTE_LOCATION_POSITION, "Position");
  glBindAttribLocation(Program, VERTEX_ATTRIBUTE_LOCATION_NORMAL, "Normal");
  glBindAttribLocation(Program, VERTEX_ATTRIBUTE_LOCATION_TANGENT, "Tangent");
  glBindAttribLocation(Program, VERTEX_ATTRIBUTE_LOCATION_BINORMAL, "Binormal");
  glBindAttribLocation(Program, VERTEX_ATTRIBUTE_LOCATION_COLOR, "VertexColor");
  glBindAttribLocation(Program, VERTEX_ATTRIBUTE_LOCATION_UV0, "TexCoord");
  glBindAttribLocation(Program, VERTEX_ATTRIBUTE_LOCATION_UV1, "TexCoord1");
  glBindAttribLocation(Program, VERTEX_ATTRIBUTE_LOCATION_JOINT_INDICES, "JointIndices");
  glBindAttribLocation(Program, VERTEX_ATTRIBUTE_LOCATION_JOINT_WEIGHTS, "JointWeights");
  glBindAttribLocation(Program, VERTEX_ATTRIBUTE_LOCATION_FONT_PARMS, "FontParms");
  glBindAttribLocation(Program, VERTEX_ATTRIBUTE_LOCATION_INDICES, "Indices");

  //--------------------------
  // Link Program
  //--------------------------

  glLinkProgram(Program);

  GLint linkStatus;
  glGetProgramiv(Program, GL_LINK_STATUS, &linkStatus);
  if (linkStatus == GL_FALSE) {
    GLchar msg[1024];
    glGetProgramInfoLog(Program, sizeof(msg), 0, msg);
    Free();
    OVR_LOG("Linking program failed: %s\n", msg);
    OVR_FAIL("Failed to link program");
    return;
  }

  //--------------------------
  // Determine Uniform Parm Location and Binding.
  //--------------------------

  numTextureBindings = 0;
  numUniformBufferBindings = 0;

  // Globally-defined system level uniforms.
  {
    ViewID.Type = ovrProgramParmType::INT;
    ViewID.Location = glGetUniformLocation(Program, "ViewID");
    ViewID.Binding = ViewID.Location;

    SceneMatrices.Type = ovrProgramParmType::BUFFER_UNIFORM;
    SceneMatrices.Location = glGetUniformBlockIndex(Program, "SceneMatrices");
    SceneMatrices.Binding = numUniformBufferBindings++;
    glUniformBlockBinding(Program, SceneMatrices.Location, SceneMatrices.Binding);

    ModelMatrix.Type = ovrProgramParmType::FLOAT_MATRIX4;
    ModelMatrix.Location = glGetUniformLocation(Program, "ModelMatrix");
    ModelMatrix.Binding = ModelMatrix.Location;
  }

  glUseProgram(Program);

  for (int i = 0; i < numParms; ++i) {
    assert(parms[i].Type != ovrProgramParmType::MAX);
    Uniforms[i].Type = parms[i].Type;
    if (parms[i].Type == ovrProgramParmType::TEXTURE_SAMPLED) {
      Uniforms[i].Location = static_cast<int16_t>(glGetUniformLocation(Program, parms[i].Name));
      Uniforms[i].Binding = numTextureBindings++;
      glUniform1i(Uniforms[i].Location, Uniforms[i].Binding);
    } else if (parms[i].Type == ovrProgramParmType::BUFFER_UNIFORM) {
      Uniforms[i].Location = glGetUniformBlockIndex(Program, parms[i].Name);
      Uniforms[i].Binding = numUniformBufferBindings++;
      glUniformBlockBinding(Program, Uniforms[i].Location, Uniforms[i].Binding);
    } else if (parms[i].Type == ovrProgramParmType::BUFFER_SSBO) {
      OVR_FAIL(
          "Shader storage buffers should only be used for compute shaders, Exynos(Gear) limitation")
    } else {
      Uniforms[i].Location = static_cast<int16_t>(glGetUniformLocation(Program, parms[i].Name));
      Uniforms[i].Binding = Uniforms[i].Location;
    }

    if (Uniforms[i].Location < 0 || Uniforms[i].Binding < 0) {
      OVR_WARN("ovrGlProgram::Build. Invalid shader parm: %s", parms[i].Name);
      assert(false);
    }
  }

  glUseProgram(0);
}

ovrGlProgram::ovrGlProgram(
    const char* computeDirectives,
    const char* computeSrc,
    const ovrProgramParm* parms,
    const int numParms) {
  assert(numParms <= 12);

  ComputeShader = CompileShader(GL_COMPUTE_SHADER, computeDirectives, computeSrc);
  if (ComputeShader == 0) {
    Free();
    OVR_FAIL("Failed to compile compute shader");
    return;
  }

  Program = glCreateProgram();
  glAttachShader(Program, ComputeShader);

  //--------------------------
  // Link Program
  //--------------------------

  glLinkProgram(Program);

  GLint linkStatus;
  glGetProgramiv(Program, GL_LINK_STATUS, &linkStatus);
  if (linkStatus == GL_FALSE) {
    GLchar msg[1024];
    glGetProgramInfoLog(Program, sizeof(msg), 0, msg);
    Free();
    OVR_LOG("Linking program failed: %s\n", msg);
    OVR_FAIL("Failed to link program");
    return;
  }

  //--------------------------
  // Determine Uniform Parm Location and Binding.
  //--------------------------

  numTextureBindings = 0;
  numUniformBufferBindings = 0;

  glUseProgram(Program);

  for (int i = 0; i < numParms; ++i) {
    assert(parms[i].Type != ovrProgramParmType::MAX);
    Uniforms[i].Type = parms[i].Type;
    if (parms[i].Type == ovrProgramParmType::TEXTURE_SAMPLED) {
      Uniforms[i].Location = static_cast<int16_t>(glGetUniformLocation(Program, parms[i].Name));
      Uniforms[i].Binding = numTextureBindings++;
      glUniform1i(Uniforms[i].Location, Uniforms[i].Binding);
    } else if (parms[i].Type == ovrProgramParmType::BUFFER_UNIFORM) {
      Uniforms[i].Location = glGetUniformBlockIndex(Program, parms[i].Name);
      Uniforms[i].Binding = numUniformBufferBindings++;
      glUniformBlockBinding(Program, Uniforms[i].Location, Uniforms[i].Binding);
    } else if (parms[i].Type == ovrProgramParmType::BUFFER_SSBO) {
      Uniforms[i].Location =
          glGetProgramResourceIndex(Program, GL_SHADER_STORAGE_BLOCK, parms[i].Name);
      // shader storage buffers only support newer shader specified binding location.
      // should change everything to that.
      Uniforms[i].Binding = -1;
    } else {
      Uniforms[i].Location = static_cast<int16_t>(glGetUniformLocation(Program, parms[i].Name));
      Uniforms[i].Binding = Uniforms[i].Location;
    }

    if (Uniforms[i].Location < 0 ||
        (Uniforms[i].Binding < 0 && parms[i].Type != ovrProgramParmType::BUFFER_SSBO)) {
      OVR_WARN("ovrGlProgram::Build. Invalid shader parm: %s", parms[i].Name);
      assert(false);
    }
  }

  glUseProgram(0);
}

ovrGlProgram::ovrGlProgram(ovrGlProgram&& other) {
  *this = move(other);
}

ovrGlProgram& ovrGlProgram::operator=(ovrGlProgram&& other) {
  Free();

  memcpy(this, &other, sizeof(ovrGlProgram));
  other.Program = 0;
  other.FragmentShader = 0;
  other.VertexShader = 0;
  other.ComputeShader = 0;

  return *this;
}

ovrGlProgram::~ovrGlProgram() {
  Free();
}

void ovrGlProgram::Free() {
  glUseProgram(0);
  if (Program != 0) {
    glDeleteProgram(Program);
  }
  if (VertexShader != 0) {
    glDeleteShader(VertexShader);
  }
  if (FragmentShader != 0) {
    glDeleteShader(FragmentShader);
  }
  if (ComputeShader != 0) {
    glDeleteShader(ComputeShader);
  }
  Program = 0;
  VertexShader = 0;
  FragmentShader = 0;
}

void ovrGlProgram::SetUseMultiview(bool useMultiview) {
  UseMultiview = useMultiview;
}
