#pragma once

#define MAX_JOINTS 25

namespace ovrUtilities {

enum ovrVertexAttributeLocation {
  VERTEX_ATTRIBUTE_LOCATION_POSITION = 0,
  VERTEX_ATTRIBUTE_LOCATION_NORMAL = 1,
  VERTEX_ATTRIBUTE_LOCATION_TANGENT = 2,
  VERTEX_ATTRIBUTE_LOCATION_BINORMAL = 3,
  VERTEX_ATTRIBUTE_LOCATION_COLOR = 4,
  VERTEX_ATTRIBUTE_LOCATION_UV0 = 5,
  VERTEX_ATTRIBUTE_LOCATION_UV1 = 6,
  VERTEX_ATTRIBUTE_LOCATION_JOINT_INDICES = 7,
  VERTEX_ATTRIBUTE_LOCATION_JOINT_WEIGHTS = 8,
  VERTEX_ATTRIBUTE_LOCATION_FONT_PARMS = 9,
  VERTEX_ATTRIBUTE_LOCATION_INDICES = 10
};

enum class ovrProgramParmType : char {
  INT, // int
  INT_VECTOR2, // Vector2i
  INT_VECTOR3, // Vector3i
  INT_VECTOR4, // Vector4i
  FLOAT, // float
  FLOAT_VECTOR2, // Vector2f
  FLOAT_VECTOR3, // Vector3f
  FLOAT_VECTOR3x2, // Vector3f[2]
  FLOAT_VECTOR4, // Vector4f
  FLOAT_MATRIX4, // Matrix4f (always inverted for GL)
  TEXTURE_SAMPLED, // GlTexture
  BUFFER_UNIFORM, // read-only uniform buffer (GLSL: uniform)
  BUFFER_SSBO, // Shader storage buffer
  MAX
};

struct ovrProgramParm {
  const char* Name;
  ovrProgramParmType Type;
};

struct ovrUniform {
  static const int MAX_UNIFORMS = 12;

  int Location{-1}; // the index of the uniform in the render program
  int Binding{-1}; // the resource binding (eg. texture image unit or uniform block)
  ovrProgramParmType Type{ovrProgramParmType::MAX}; // the type of the data
};

struct ovrGlProgram {
  ovrGlProgram() = default;
  // Create vertex/fragment program
  ovrGlProgram(
      const char* vertexDirectives,
      const char* vertexSrc,
      const char* fragmentDirectives,
      const char* fragmentSrc,
      const ovrProgramParm* parms,
      const int numParms);
  // Create compute shader
  ovrGlProgram(
      const char* computeDirectives,
      const char* computeSrc,
      const ovrProgramParm* parms,
      const int numParms);
  ~ovrGlProgram();
  ovrGlProgram(const ovrGlProgram&) = delete;
  ovrGlProgram(ovrGlProgram&& other);
  ovrGlProgram& operator=(const ovrGlProgram&) = delete;
  ovrGlProgram& operator=(ovrGlProgram&& other);

  // Must be set before the creation of any GlPrograms
  static void SetUseMultiview(bool useMultiview);

  bool IsValid() const;

  static const int MAX_VIEWS = 2;
  static const int SCENE_MATRICES_UBO_SIZE = 2 * 16 * MAX_VIEWS;

  unsigned int Program{0};
  unsigned int VertexShader{0};
  unsigned int FragmentShader{0};
  unsigned int ComputeShader{0};

  // Globally-defined system level uniforms.
  ovrUniform ViewID; // uniform for ViewID; is -1 if OVR_multiview unavailable or disabled
  ovrUniform ModelMatrix; // uniform for "uniform mat4 ModelMatrix;"
  ovrUniform SceneMatrices; // uniform for "SceneMatrices" ubo :
                            // uniform SceneMatrices {
                            //   mat4 ViewMatrix[NUM_VIEWS];
                            //   mat4 ProjectionMatrix[NUM_VIEWS];
                            // } sm;

  ovrUniform Uniforms[ovrUniform::MAX_UNIFORMS];
  int numTextureBindings{0};
  int numUniformBufferBindings{0};

 private:
  void Free();
};

inline bool ovrGlProgram::IsValid() const {
  return Program != 0;
}

} // namespace ovrUtilities
