#include "GlTexture.h"

#include <algorithm>
#include <cassert>
#include <cmath>
#include "Log.h"

using namespace std;
using namespace ovrUtilities;

ovrGlTexture::ovrGlTexture(uint32_t texture, uint32_t target) : Texture(texture), Target(target) {}

ovrGlTexture::ovrGlTexture(
    uint32_t target,
    uint32_t format,
    uint32_t width,
    uint32_t height,
    uint32_t mips) {
  assert(target == GL_TEXTURE_2D || target == GL_TEXTURE_CUBE_MAP);
  Target = target;

  glGenTextures(1, &Texture);
  glBindTexture(Target, Texture);
  glTexStorage2D(Target, mips, format, width, height);
  glBindTexture(Target, 0);
}

ovrGlTexture::ovrGlTexture(
    uint32_t target,
    uint32_t format,
    uint32_t width,
    uint32_t height,
    uint32_t depth,
    uint32_t mips) {
  assert(target == GL_TEXTURE_2D_ARRAY);
  Target = target;

  glGenTextures(1, &Texture);
  glBindTexture(Target, Texture);
  glTexStorage3D(Target, mips, format, width, height, depth);
  glBindTexture(Target, 0);
}

ovrGlTexture::ovrGlTexture(ovrGlBuffer& buffer, uint32_t format) {
  Target = GL_TEXTURE_BUFFER_EXT;

  glGenTextures(1, &Texture);
  glBindTexture(Target, Texture);
  glTexBufferEXT(Target, format, buffer.GetGlID());
  glBindTexture(Target, 0);
}

ovrGlTexture::~ovrGlTexture() {
  Free();
}

ovrGlTexture::ovrGlTexture(ovrGlTexture&& other) {
  *this = move(other);
}

ovrGlTexture& ovrGlTexture::operator=(ovrGlTexture&& other) {
  Free();

  Texture = other.Texture;
  Target = other.Target;

  other.Texture = 0;

  return *this;
}

void ovrGlTexture::Free() {
  if (Texture != 0) {
    glDeleteTextures(1, &Texture);
    Texture = 0;
  }
}

void ovrGlTexture::Activate(uint32_t binding) {
  glActiveTexture(GL_TEXTURE0 + binding);
  glBindTexture(Target, Texture);
}

void ovrGlTexture::DeActivate(uint32_t binding) {
  glActiveTexture(GL_TEXTURE0 + binding);
  glBindTexture(Target, 0);
}

/*
This taken(including the rest of this comment) from app frameworks GlTexture. Modified to produce an
ovrUtilities GlTexture instead

KTX Container Format

KTX is a format for storing textures for OpenGL and OpenGL ES applications.
It is distinguished by the simplicity of the loader required to instantiate
a GL texture object from the file contents.

Byte[12] identifier
UInt32 endianness
UInt32 glType
UInt32 glTypeSize
UInt32 glFormat
Uint32 glInternalFormat
Uint32 glBaseInternalFormat
UInt32 pixelWidth
UInt32 pixelHeight
UInt32 pixelDepth
UInt32 numberOfArrayElements
UInt32 numberOfFaces
UInt32 numberOfMipmapLevels
UInt32 bytesOfKeyValueData

for each keyValuePair that fits in bytesOfKeyValueData
    UInt32   keyAndValueByteSize
    Byte     keyAndValue[keyAndValueByteSize]
    Byte     valuePadding[3 - ((keyAndValueByteSize + 3) % 4)]
end

for each mipmap_level in numberOfMipmapLevels*
    UInt32 imageSize;
    for each array_element in numberOfArrayElements*
       for each face in numberOfFaces
           for each z_slice in pixelDepth*
               for each row or row_of_blocks in pixelHeight*
                   for each pixel or block_of_pixels in pixelWidth
                       Byte data[format-specific-number-of-bytes]**
                   end
               end
           end
           Byte cubePadding[0-3]
       end
    end
    Byte mipPadding[3 - ((imageSize + 3) % 4)]
end

*/

namespace {
#pragma pack(1)
struct OVR_KTX_HEADER {
  uint8_t identifier[12];
  uint32_t endianness;
  uint32_t glType;
  uint32_t glTypeSize;
  uint32_t glFormat;
  uint32_t glInternalFormat;
  uint32_t glBaseInternalFormat;
  uint32_t pixelWidth;
  uint32_t pixelHeight;
  uint32_t pixelDepth;
  uint32_t numberOfArrayElements;
  uint32_t numberOfFaces;
  uint32_t numberOfMipmapLevels;
  uint32_t bytesOfKeyValueData;
};
#pragma pack()
} // namespace

ovrTextureInfo ovrUtilities::GetKTXTextureInfo(const uint8_t* data, size_t dataSize) {
  ovrTextureInfo info;

  if (dataSize < (sizeof(OVR_KTX_HEADER))) {
    OVR_WARN("Invalid KTX file, not large enough to even hold the header");
    return ovrTextureInfo{};
  }

  const char fileIdentifier[12] = {
      '\xAB', 'K', 'T', 'X', ' ', '1', '1', '\xBB', '\r', '\n', '\x1A', '\n'};

  const OVR_KTX_HEADER& header = *(OVR_KTX_HEADER*)data;
  if (memcmp(header.identifier, fileIdentifier, sizeof(fileIdentifier)) != 0) {
    OVR_WARN("Invalid KTX file");
    return ovrTextureInfo{};
  }
  // only support little endian
  if (header.endianness != 0x04030201) {
    OVR_WARN("KTX file has wrong endianess");
    return ovrTextureInfo{};
  }
  // only support compressed or unsigned byte
  if (header.glType != 0 && header.glType != GL_UNSIGNED_BYTE) {
    OVR_WARN("KTX file has unsupported glType %d", header.glType);
    return ovrTextureInfo{};
  }
  // no support for texture arrays
  if (header.numberOfArrayElements != 0) {
    OVR_WARN("KTX file has unsupported number of array elements %d", header.numberOfArrayElements);
    return ovrTextureInfo{};
  }

  if (!(header.numberOfFaces == 1 || header.numberOfFaces == 6)) {
    OVR_WARN("KTX file must have one or 6 faces");
    return ovrTextureInfo{};
  }

  // skip the key value data
  const size_t startTex = sizeof(OVR_KTX_HEADER) + header.bytesOfKeyValueData;
  if ((startTex < sizeof(OVR_KTX_HEADER)) || (startTex >= dataSize)) {
    OVR_WARN("Invalid KTX header sizes, bytesOfKeyValueData seems to be garbage");
    return ovrTextureInfo{};
  }

  if (header.pixelWidth <= 0 || header.pixelWidth > 32768 || header.pixelHeight <= 0 ||
      header.pixelHeight > 32768) {
    OVR_LOG("Invalid texture size (%dx%d)", header.pixelWidth, header.pixelHeight);
    return ovrTextureInfo{};
  }

  if (header.numberOfMipmapLevels <= 0) {
    OVR_LOG("Invalid mip count %d", header.numberOfMipmapLevels);
    return ovrTextureInfo{};
  }

  info.Width = header.pixelWidth;
  info.Height = header.pixelHeight;
  info.NumMips = header.numberOfMipmapLevels;
  info.NumFaces = header.numberOfFaces;
  info.GlFormat = header.glFormat;
  info.GlInternalFormat = header.glInternalFormat;
  info.OffsetToData = startTex;

  return info;
}

void ovrUtilities::LoadKTXTexture(
    const uint8_t* data,
    size_t dataSize,
    const ovrTextureInfo& info) {
  const unsigned char* level = (const unsigned char*)data + info.OffsetToData;
  const unsigned char* endOfBuffer = data + dataSize;

  auto target = info.NumFaces <= 1 ? GL_TEXTURE_2D : GL_TEXTURE_CUBE_MAP;

  int w = info.Width;
  int h = info.Height;
  for (uint32_t i = 0; i < info.NumMips; i++) {
    int32_t mipSize = static_cast<int32_t>(*(const size_t*)level);

    level += 4;
    if (level > endOfBuffer) {
      OVR_WARN("Image data exceeds buffer size");
      return;
    }

    if (mipSize <= 0 || mipSize > endOfBuffer - level) {
      OVR_WARN(
          "Mip level %d exceeds buffer size (%d > %td)",
          i,
          mipSize,
          ptrdiff_t(endOfBuffer - level));
      return;
    }

    for (uint32_t face = 0; face < info.NumFaces; ++face) {
      auto realTarget = target;
      if (info.NumFaces == 6) {
        realTarget = GL_TEXTURE_CUBE_MAP_POSITIVE_X + face;
      }

      if (info.GlFormat == 0) {
        glCompressedTexSubImage2D(realTarget, i, 0, 0, w, h, info.GlInternalFormat, mipSize, level);
      } else {
        glTexSubImage2D(realTarget, i, 0, 0, w, h, info.GlFormat, GL_UNSIGNED_BYTE, level);
      }
      level += mipSize;
      level += 3 - ((mipSize + 3) % 4);
    }

    if (level > endOfBuffer) {
      OVR_WARN("Image data exceeds buffer size");
      return;
    }

    w >>= 1;
    h >>= 1;
    if (w < 1) {
      w = 1;
    }
    if (h < 1) {
      h = 1;
    }
  }

  glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_REPEAT);
  if (info.NumMips <= 1) {
    glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  } else {
    glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
  }
  glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}

ovrGlTexture ovrUtilities::LoadKTXTexture(const uint8_t* data, size_t dataSize) {
  ovrTextureInfo info = GetKTXTextureInfo(data, dataSize);

  ovrGlTexture result{(uint32_t)(info.NumFaces <= 1 ? GL_TEXTURE_2D : GL_TEXTURE_CUBE_MAP),
                      info.GlInternalFormat,
                      info.Width,
                      info.Height,
                      info.NumMips};

  result.Bind([&](auto target) { LoadKTXTexture(data, dataSize, info); });

  return result;
}
