#pragma once

#include <functional>
#include <mutex>
#include <shared_mutex>

namespace ovrUtilities {
template <typename T>
class Locked {
 public:
  Locked() = default;
  ~Locked() = default;
  Locked(const Locked& other) {
    std::shared_lock<std::shared_timed_mutex> lockOther(other.Mutex);
    Object(other.Object);
  }
  Locked& operator=(const Locked& other) {
    std::unique_lock<std::shared_timed_mutex> lock(Mutex);
    std::shared_lock<std::shared_timed_mutex> lockOther(other.Mutex);
    Object = other.Object;
  }
  Locked(Locked&& other) {
    std::unique_lock<std::shared_timed_mutex> lockOther(other.Mutex);
    Object = std::move(other.Object);
  }
  Locked& operator=(Locked&& other) {
    std::unique_lock<std::shared_timed_mutex> lock1(Mutex, std::defer_lock);
    std::unique_lock<std::shared_timed_mutex> lock2(other.Mutex, std::defer_lock);
    std::lock(lock1, lock2);
    Object = std::move(other.Object);
  }

  Locked(const T& arg) {
    Object = arg;
  }

  Locked(T&& arg) {
    Object = arg;
  }

  Locked& operator=(const T& arg) {
    std::unique_lock<std::shared_timed_mutex> lock(Mutex);
    Object = arg;
  }

  Locked& operator=(T&& arg) {
    std::unique_lock<std::shared_timed_mutex> lock(Mutex);
    Object = std::move(arg);
  }

  friend bool operator==(const Locked& arg1, const Locked& arg2) {
    std::shared_lock<std::shared_timed_mutex> lock1(arg1.Mutex, std::defer_lock);
    std::shared_lock<std::shared_timed_mutex> lock2(arg2.Mutex, std::defer_lock);
    std::lock(lock1, lock2);
    return arg1.Object == arg2.Object;
  }

  friend bool operator!=(const Locked& arg1, const Locked& arg2) {
    return !(arg1 == arg2);
  }

  friend bool operator<(const Locked& arg1, const Locked& arg2) {
    std::shared_lock<std::shared_timed_mutex> lock1(arg1.Mutex, std::defer_lock);
    std::shared_lock<std::shared_timed_mutex> lock2(arg2.Mutex, std::defer_lock);
    std::lock(lock1, lock2);
    return arg1.Object < arg2.Object;
  }

  friend bool operator>(const Locked& arg1, const Locked& arg2) {
    std::shared_lock<std::shared_timed_mutex> lock1(arg1.Mutex, std::defer_lock);
    std::shared_lock<std::shared_timed_mutex> lock2(arg2.Mutex, std::defer_lock);
    std::lock(lock1, lock2);
    return !(arg1.Object > arg2.Object);
  }

  friend bool operator<=(const Locked& arg1, const Locked& arg2) {
    return !(arg1 > arg2);
  }

  friend bool operator>=(const Locked& arg1, const Locked& arg2) {
    return !(arg1 < arg2);
  }

  template <typename OperationType>
  auto operator()(OperationType operation) const {
    std::shared_lock<std::shared_timed_mutex> lock(Mutex);
    return operation(Object);
  }

  template <typename OperationType>
  auto operator()(OperationType operation) {
    std::unique_lock<std::shared_timed_mutex> lock(Mutex);
    return operation(Object);
  }

 private:
  T Object;
  mutable std::shared_timed_mutex Mutex;
};
} // namespace ovrUtilities
