#include "../AudioEngine.h"
#include <Audioclient.h>
#include <Mmdeviceapi.h>
#include <Windows.h>
#include <atlbase.h>
#include <atomic>
#include <cassert>
#include <thread>
#include "../Utilities/Log.h"

namespace {
PlayCallback_t Callback;

std::thread PlayThread;
std::atomic_bool ShutdownPlayThread;

// taken from old windows stadium client. not currently used. useful if you want to play audio on
// rift
CComPtr<IAudioClient> CreateAudioClient(WCHAR* riftDeviceId, bool microphone) {
  CComPtr<IMMDeviceEnumerator> enumerator;
  auto comResult = CoCreateInstance(
      __uuidof(MMDeviceEnumerator),
      nullptr,
      CLSCTX_ALL,
      __uuidof(IMMDeviceEnumerator),
      (void**)&enumerator);
  if (FAILED(comResult)) {
    OVR_FAIL("Failed to create devices enumerator, hresult %d", comResult);
  }

  CComPtr<IMMDeviceCollection> devices;
  comResult = enumerator->EnumAudioEndpoints(
      microphone ? eCapture : eRender, DEVICE_STATE_ACTIVE, &devices);
  if (FAILED(comResult)) {
    OVR_FAIL("Failed to enumerate devices, hresult %d", comResult);
  }

  uint32_t deviceCount;
  comResult = devices->GetCount(&deviceCount);
  if (FAILED(comResult)) {
    OVR_FAIL("Failed to get device count, hresult %d", comResult);
  }

  for (uint32_t i = 0; i < deviceCount; ++i) {
    CComPtr<IMMDevice> device;
    wchar_t* deviceId;
    devices->Item(i, &device);
    comResult = device->GetId(&deviceId);
    if (FAILED(comResult)) {
      OVR_FAIL("Failed to device id, hresult %d", comResult);
    }
    // OutputDebugStringW( L"enumerated Device Id:" );
    // OutputDebugStringW( deviceId );
    // OutputDebugStringW( L"\n" );
    if (wcscmp(riftDeviceId, deviceId) == 0) {
      CComPtr<IAudioClient> result;
      comResult = device->Activate(__uuidof(IAudioClient), CLSCTX_ALL, nullptr, (void**)&result);
      if (FAILED(comResult)) {
        OVR_FAIL("Failed to activate rift mic device, hresult %d", comResult);
      }
      CoTaskMemFree(deviceId);
      return result;
      break;
    }
    CoTaskMemFree(deviceId);
  }
  return nullptr;
}

CComPtr<IAudioClient> CreateDefaultAudioClient() {
  CComPtr<IMMDeviceEnumerator> enumerator;
  auto comResult = CoCreateInstance(
      __uuidof(MMDeviceEnumerator),
      nullptr,
      CLSCTX_ALL,
      __uuidof(IMMDeviceEnumerator),
      (void**)&enumerator);
  if (FAILED(comResult)) {
    OVR_FAIL("Failed to create devices enumerator, hresult %d", comResult);
  }

  CComPtr<IMMDevice> device;
  enumerator->GetDefaultAudioEndpoint(eRender, ERole::eConsole, &device);

  CComPtr<IAudioClient> result;
  comResult = device->Activate(__uuidof(IAudioClient), CLSCTX_ALL, nullptr, (void**)&result);
  if (FAILED(comResult)) {
    OVR_FAIL("Failed to activate rift mic device, hresult %d", comResult);
  }
  return result;
}

WAVEFORMATEX SetupMixFormat(CComPtr<IAudioClient>& audioClient) {
  WAVEFORMATEX* deviceFormat = nullptr;
  auto comResult = audioClient->GetMixFormat(&deviceFormat);
  if (FAILED(comResult)) {
    OVR_FAIL("Failed to get device format, hresult %d", comResult);
  }

  WAVEFORMATEX requestedFormat = *deviceFormat;
  requestedFormat.cbSize = 0;
  requestedFormat.wBitsPerSample = 16;
  requestedFormat.nSamplesPerSec = SampleRate;
  requestedFormat.nBlockAlign = 2 * requestedFormat.nChannels;
  requestedFormat.nAvgBytesPerSec = requestedFormat.nSamplesPerSec * requestedFormat.nBlockAlign;
  requestedFormat.wFormatTag = WAVE_FORMAT_PCM;

  CoTaskMemFree(deviceFormat);
  deviceFormat = nullptr;
  comResult =
      audioClient->IsFormatSupported(AUDCLNT_SHAREMODE_SHARED, &requestedFormat, &deviceFormat);
  CoTaskMemFree(deviceFormat);
  if (FAILED(comResult)) {
    OVR_FAIL("Failed to set device format to 16bit samples, hresult %x", comResult);
  }
  return requestedFormat;
}
} // namespace

void AudioEngineInitialize(PlayCallback_t playCallback, MicCallback_t micCallback) {
  Callback = playCallback;

  ShutdownPlayThread.store(false);
  PlayThread = std::thread([]() {
    auto comResult = CoInitializeEx(nullptr, COINIT_MULTITHREADED);
    if (FAILED(comResult)) {
      assert(false);
    }
    // AUDCLNT_E_SERVICE_NOT_RUNNING
    CComPtr<IAudioClient> WASAPIAudioClient;
    CComPtr<IAudioRenderClient> WASAPIRenderClient;
    HANDLE WASAPIAudioEvent;

    WASAPIAudioClient = CreateDefaultAudioClient();

    if (!WASAPIAudioClient.p) {
      OVR_FAIL("Failed to find default audio device");
    }

    WAVEFORMATEX haudioFormat = SetupMixFormat(WASAPIAudioClient);
    OVR_LOG(
        "audio format: channels: %d samplesPerSec: %d bitsPerSample: %d",
        haudioFormat.nChannels,
        haudioFormat.nSamplesPerSec,
        haudioFormat.wBitsPerSample,
        haudioFormat.wBitsPerSample);
    /*if (haudioFormat.nChannels != 2 || haudioFormat.wBitsPerSample != 16 ||
        haudioFormat.nSamplesPerSec != SampleRate) {
      FAIL("unexpected audio format");
    }*/

    float bufferTime = ((float)SamplesPerTick / SampleRate) * 10000000.0f; // in 100's of
                                                                           // nanoseconds
    comResult = WASAPIAudioClient->Initialize(
        AUDCLNT_SHAREMODE_SHARED,
        AUDCLNT_STREAMFLAGS_EVENTCALLBACK | AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM |
            AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY,
        (REFERENCE_TIME)bufferTime,
        0,
        &haudioFormat,
        nullptr);
    if (FAILED(comResult)) {
      OVR_FAIL("Failed to initialize audio stream, hresult %x", comResult);
    }

    WASAPIAudioEvent = CreateEvent(nullptr, FALSE, FALSE, nullptr);
    if (WASAPIAudioEvent == 0) {
      OVR_FAIL("Failed to create audio event");
    }
    comResult = WASAPIAudioClient->SetEventHandle(WASAPIAudioEvent);
    if (FAILED(comResult)) {
      OVR_FAIL("Failed to set audio event, hresult %x", comResult);
    }
    comResult =
        WASAPIAudioClient->GetService(__uuidof(IAudioRenderClient), (void**)&WASAPIRenderClient);
    if (FAILED(comResult)) {
      OVR_FAIL("Failed to get audio render interface, hresult %x", comResult);
    }

    WASAPIAudioClient->Start();

    while (!ShutdownPlayThread.load()) {
      DWORD retval = WaitForSingleObject(WASAPIAudioEvent, 500);
      if (retval != WAIT_OBJECT_0) {
        // Something went wrong, no buffer in half a second
        ShutdownPlayThread.store(true);
      }
      if (!ShutdownPlayThread.load()) {
        short* renderBuf;
        comResult = WASAPIRenderClient->GetBuffer(SamplesPerTick, (BYTE**)&renderBuf);
        if (SUCCEEDED(comResult)) {
          if (Callback) {
            Callback(renderBuf);
          }

          WASAPIRenderClient->ReleaseBuffer(SamplesPerTick, 0);
        }
      }
    }

    WASAPIAudioClient->Stop();

    CoUninitialize();
  });
}

void AudioEngineShutdown() {
  if (PlayThread.joinable()) {
    ShutdownPlayThread.store(true);
    PlayThread.join();
  }
}
