/*****************************************************************************

Filename    :   main.cpp
Content     :   Simple minimal VR demo
Created     :   December 1, 2014
Author      :   Tom Heath
Copyright   :   Copyright (c) Facebook Technologies, LLC and its affiliates. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

/*****************************************************************************/
/// This sample has not yet been fully assimiliated into the framework
/// and also the GL support is not quite fully there yet, hence the VR
/// is not that great!

//#include "../../OculusRoomTiny_Advanced/Common/Win32_GLAppUtil.h"

// Include the Oculus SDK
#include "Extras/OVR_Math.h"
#include "OVR_CAPI_GL.h"

#if defined(_WIN32)
#include <dxgi.h> // for GetDefaultAdapterLuid
#pragma comment(lib, "dxgi.lib")
#endif

#include <GL/glew.h>
#include <GL/wglew.h>

#include "glm/gtc/type_ptr.hpp"

#include <cassert>
#include <filesystem>
#include <iterator>

#include "../Application.h"
#include "../Utilities/GlBuffer.h"
#include "../Utilities/GlProgram.h"

using namespace OVR;
using namespace ovrUtilities;
using namespace std;
using namespace glm;

namespace fs = std::experimental::filesystem;

#ifndef VALIDATE
#define VALIDATE(x, msg)                                      \
  if (!(x)) {                                                 \
    MessageBoxA(NULL, (msg), "Mirror", MB_ICONERROR | MB_OK); \
    exit(-1);                                                 \
  }
#endif

//-------------------------------------------------------------------------------------------
struct OGL {
  static const bool UseDebugContext = false;

  HWND Window;
  HDC hDC;
  HGLRC WglContext;
  bool Running;
  bool Key[256];
  int WinSizeW;
  int WinSizeH;
  GLuint fboId;
  HINSTANCE hInstance;

  static LRESULT CALLBACK
  WindowProc(_In_ HWND hWnd, _In_ UINT Msg, _In_ WPARAM wParam, _In_ LPARAM lParam) {
    OGL* p = reinterpret_cast<OGL*>(GetWindowLongPtr(hWnd, 0));
    switch (Msg) {
      case WM_KEYDOWN:
        p->Key[wParam] = true;
        break;
      case WM_KEYUP:
        p->Key[wParam] = false;
        break;
      case WM_DESTROY:
        p->Running = false;
        break;
      default:
        return DefWindowProcW(hWnd, Msg, wParam, lParam);
    }
    if ((p->Key['Q'] && p->Key[VK_CONTROL]) || p->Key[VK_ESCAPE]) {
      p->Running = false;
    }
    return 0;
  }

  OGL()
      : Window(nullptr),
        hDC(nullptr),
        WglContext(nullptr),
        Running(false),
        WinSizeW(0),
        WinSizeH(0),
        fboId(0),
        hInstance(nullptr) {
    // Clear input
    for (int i = 0; i < sizeof(Key) / sizeof(Key[0]); ++i)
      Key[i] = false;
  }

  ~OGL() {
    ReleaseDevice();
    CloseWindow();
  }

  bool InitWindow(HINSTANCE hInst, LPCWSTR title) {
    hInstance = hInst;
    Running = true;

    WNDCLASSW wc;
    memset(&wc, 0, sizeof(wc));
    wc.style = CS_CLASSDC;
    wc.lpfnWndProc = WindowProc;
    wc.cbWndExtra = sizeof(struct OGL*);
    wc.hInstance = GetModuleHandleW(NULL);
    wc.lpszClassName = L"ORT";
    RegisterClassW(&wc);

    // adjust the window size and show at InitDevice time
    Window =
        CreateWindowW(wc.lpszClassName, title, WS_OVERLAPPEDWINDOW, 0, 0, 0, 0, 0, 0, hInstance, 0);
    if (!Window)
      return false;

    SetWindowLongPtr(Window, 0, LONG_PTR(this));

    hDC = GetDC(Window);

    return true;
  }

  void CloseWindow() {
    if (Window) {
      if (hDC) {
        ReleaseDC(Window, hDC);
        hDC = nullptr;
      }
      DestroyWindow(Window);
      Window = nullptr;
      UnregisterClassW(L"OGL", hInstance);
    }
  }

  // Note: currently there is no way to get GL to use the passed pLuid
  bool InitDevice(int vpW, int vpH, const LUID* /*pLuid*/, bool windowed = true) {
    UNREFERENCED_PARAMETER(windowed);

    WinSizeW = vpW;
    WinSizeH = vpH;

    RECT size = {0, 0, vpW, vpH};
    AdjustWindowRect(&size, WS_OVERLAPPEDWINDOW, false);
    const UINT flags = SWP_NOMOVE | SWP_NOZORDER | SWP_SHOWWINDOW;
    if (!SetWindowPos(Window, nullptr, 0, 0, size.right - size.left, size.bottom - size.top, flags))
      return false;

    PFNWGLCHOOSEPIXELFORMATARBPROC wglChoosePixelFormatARBFunc = nullptr;
    PFNWGLCREATECONTEXTATTRIBSARBPROC wglCreateContextAttribsARBFunc = nullptr;
    {
      // First create a context for the purpose of getting access to wglChoosePixelFormatARB /
      // wglCreateContextAttribsARB.
      PIXELFORMATDESCRIPTOR pfd;
      memset(&pfd, 0, sizeof(pfd));
      pfd.nSize = sizeof(pfd);
      pfd.nVersion = 1;
      pfd.iPixelType = PFD_TYPE_RGBA;
      pfd.dwFlags = PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW | PFD_DOUBLEBUFFER;
      pfd.cColorBits = 32;
      pfd.cDepthBits = 16;
      int pf = ChoosePixelFormat(hDC, &pfd);
      VALIDATE(pf, "Failed to choose pixel format.");

      VALIDATE(SetPixelFormat(hDC, pf, &pfd), "Failed to set pixel format.");

      HGLRC context = wglCreateContext(hDC);
      VALIDATE(context, "wglCreateContextfailed.");
      VALIDATE(wglMakeCurrent(hDC, context), "wglMakeCurrent failed.");

      wglChoosePixelFormatARBFunc =
          (PFNWGLCHOOSEPIXELFORMATARBPROC)wglGetProcAddress("wglChoosePixelFormatARB");
      wglCreateContextAttribsARBFunc =
          (PFNWGLCREATECONTEXTATTRIBSARBPROC)wglGetProcAddress("wglCreateContextAttribsARB");
      assert(wglChoosePixelFormatARBFunc && wglCreateContextAttribsARBFunc);

      wglDeleteContext(context);
    }

    // Now create the real context that we will be using.
    int iAttributes[] = {// WGL_DRAW_TO_WINDOW_ARB, GL_TRUE,
                         WGL_SUPPORT_OPENGL_ARB,
                         GL_TRUE,
                         WGL_COLOR_BITS_ARB,
                         32,
                         WGL_DEPTH_BITS_ARB,
                         16,
                         WGL_DOUBLE_BUFFER_ARB,
                         GL_TRUE,
                         WGL_FRAMEBUFFER_SRGB_CAPABLE_ARB,
                         GL_TRUE,
                         0,
                         0};

    float fAttributes[] = {0, 0};
    int pf = 0;
    UINT numFormats = 0;

    VALIDATE(
        wglChoosePixelFormatARBFunc(hDC, iAttributes, fAttributes, 1, &pf, &numFormats),
        "wglChoosePixelFormatARBFunc failed.");

    PIXELFORMATDESCRIPTOR pfd;
    memset(&pfd, 0, sizeof(pfd));
    VALIDATE(SetPixelFormat(hDC, pf, &pfd), "SetPixelFormat failed.");

    GLint attribs[16];
    int attribCount = 0;
    if (UseDebugContext) {
      attribs[attribCount++] = WGL_CONTEXT_FLAGS_ARB;
      attribs[attribCount++] = WGL_CONTEXT_DEBUG_BIT_ARB;
    }

    attribs[attribCount] = 0;

    WglContext = wglCreateContextAttribsARBFunc(hDC, 0, attribs);
    VALIDATE(wglMakeCurrent(hDC, WglContext), "wglMakeCurrent failed.");

    GLenum err = glewInit();
    auto foo = glewGetErrorString(err);
    VALIDATE(GLEW_OK == err, "Failed to initialize glew");

    glGenFramebuffers(1, &fboId);

    glEnable(GL_DEPTH_TEST);
    glFrontFace(GL_CW);
    glEnable(GL_CULL_FACE);

    if (UseDebugContext && GLEW_ARB_debug_output) {
      glDebugMessageCallbackARB(DebugGLCallback, NULL);
      if (glGetError()) {
        VALIDATE(false, "glDebugMessageCallbackARB failed.");
      }

      glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS_ARB);

      // Explicitly disable notification severity output.
      glDebugMessageControlARB(
          GL_DEBUG_SOURCE_API, GL_DONT_CARE, GL_DEBUG_SEVERITY_NOTIFICATION, 0, NULL, GL_FALSE);
    }

    return true;
  }

  bool HandleMessages(void) {
    MSG msg;
    while (PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE)) {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
    return Running;
  }

  void Run(bool (*MainLoop)(bool retryCreate)) {
    while (HandleMessages()) {
      // true => we'll attempt to retry for ovrError_DisplayLost
      if (!MainLoop(true))
        break;
      // Sleep a bit before retrying to reduce CPU load while the HMD is disconnected
      Sleep(10);
    }
  }

  void ReleaseDevice() {
    if (fboId) {
      glDeleteFramebuffers(1, &fboId);
      fboId = 0;
    }
    if (WglContext) {
      wglMakeCurrent(NULL, NULL);
      wglDeleteContext(WglContext);
      WglContext = nullptr;
    }
  }

  static void GLAPIENTRY DebugGLCallback(
      GLenum source,
      GLenum type,
      GLuint id,
      GLenum severity,
      GLsizei length,
      const GLchar* message,
      const void* userParam) {
    UNREFERENCED_PARAMETER(source);
    UNREFERENCED_PARAMETER(type);
    UNREFERENCED_PARAMETER(id);
    UNREFERENCED_PARAMETER(severity);
    UNREFERENCED_PARAMETER(length);
    UNREFERENCED_PARAMETER(message);
    UNREFERENCED_PARAMETER(userParam);
    char temp[1024];
    sprintf_s(temp, "Message from OpenGL: %s\n", message);
    OutputDebugStringA(temp);
  }
};

// Global OpenGL state
static OGL Platform;

struct OculusTextureBuffer {
  ovrSession Session;
  ovrTextureSwapChain ColorTextureChain;
  ovrTextureSwapChain DepthTextureChain;
  GLuint fboId;
  Sizei texSize;

  OculusTextureBuffer(ovrSession session, Sizei size, int sampleCount)
      : Session(session),
        ColorTextureChain(nullptr),
        DepthTextureChain(nullptr),
        fboId(0),
        texSize(0, 0) {
    assert(sampleCount <= 1); // The code doesn't currently handle MSAA textures.

    texSize = size;

    // This texture isn't necessarily going to be a rendertarget, but it usually is.
    assert(session); // No HMD? A little odd.

    ovrTextureSwapChainDesc desc = {};
    desc.Type = ovrTexture_2D;
    desc.ArraySize = 1;
    desc.Width = size.w;
    desc.Height = size.h;
    desc.MipLevels = 1;
    desc.Format = OVR_FORMAT_R8G8B8A8_UNORM_SRGB;
    desc.SampleCount = sampleCount;
    desc.StaticImage = ovrFalse;

    {
      ovrResult result = ovr_CreateTextureSwapChainGL(Session, &desc, &ColorTextureChain);

      int length = 0;
      ovr_GetTextureSwapChainLength(session, ColorTextureChain, &length);

      if (OVR_SUCCESS(result)) {
        for (int i = 0; i < length; ++i) {
          GLuint chainTexId;
          ovr_GetTextureSwapChainBufferGL(Session, ColorTextureChain, i, &chainTexId);
          glBindTexture(GL_TEXTURE_2D, chainTexId);

          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        }
      }
    }

    desc.Format = OVR_FORMAT_D32_FLOAT;

    {
      ovrResult result = ovr_CreateTextureSwapChainGL(Session, &desc, &DepthTextureChain);

      int length = 0;
      ovr_GetTextureSwapChainLength(session, DepthTextureChain, &length);

      if (OVR_SUCCESS(result)) {
        for (int i = 0; i < length; ++i) {
          GLuint chainTexId;
          ovr_GetTextureSwapChainBufferGL(Session, DepthTextureChain, i, &chainTexId);
          glBindTexture(GL_TEXTURE_2D, chainTexId);

          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        }
      }
    }

    glGenFramebuffers(1, &fboId);
  }

  ~OculusTextureBuffer() {
    if (ColorTextureChain) {
      ovr_DestroyTextureSwapChain(Session, ColorTextureChain);
      ColorTextureChain = nullptr;
    }
    if (DepthTextureChain) {
      ovr_DestroyTextureSwapChain(Session, DepthTextureChain);
      DepthTextureChain = nullptr;
    }
    if (fboId) {
      glDeleteFramebuffers(1, &fboId);
      fboId = 0;
    }
  }

  Sizei GetSize() const {
    return texSize;
  }

  void SetAndClearRenderSurface() {
    GLuint curColorTexId;
    GLuint curDepthTexId;
    {
      int curIndex;
      ovr_GetTextureSwapChainCurrentIndex(Session, ColorTextureChain, &curIndex);
      ovr_GetTextureSwapChainBufferGL(Session, ColorTextureChain, curIndex, &curColorTexId);
    }
    {
      int curIndex;
      ovr_GetTextureSwapChainCurrentIndex(Session, DepthTextureChain, &curIndex);
      ovr_GetTextureSwapChainBufferGL(Session, DepthTextureChain, curIndex, &curDepthTexId);
    }

    glBindFramebuffer(GL_FRAMEBUFFER, fboId);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, curColorTexId, 0);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, curDepthTexId, 0);

    glViewport(0, 0, texSize.w, texSize.h);
    glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_FRAMEBUFFER_SRGB);
  }

  void UnsetRenderSurface() {
    glBindFramebuffer(GL_FRAMEBUFFER, fboId);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, 0, 0);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, 0, 0);
  }

  void Commit() {
    ovr_CommitTextureSwapChain(Session, ColorTextureChain);
    ovr_CommitTextureSwapChain(Session, DepthTextureChain);
  }
};

static ovrGraphicsLuid GetDefaultAdapterLuid() {
  ovrGraphicsLuid luid = ovrGraphicsLuid();

#if defined(_WIN32)
  IDXGIFactory* factory = nullptr;

  if (SUCCEEDED(CreateDXGIFactory(IID_PPV_ARGS(&factory)))) {
    IDXGIAdapter* adapter = nullptr;

    if (SUCCEEDED(factory->EnumAdapters(0, &adapter))) {
      DXGI_ADAPTER_DESC desc;

      adapter->GetDesc(&desc);
      memcpy(&luid, &desc.AdapterLuid, sizeof(luid));
      adapter->Release();
    }

    factory->Release();
  }
#endif

  return luid;
}

static int Compare(const ovrGraphicsLuid& lhs, const ovrGraphicsLuid& rhs) {
  return memcmp(&lhs, &rhs, sizeof(ovrGraphicsLuid));
}

// return true to retry later (e.g. after display lost)
static bool MainLoop(bool retryCreate) {
  OculusTextureBuffer* eyeRenderTexture[2] = {nullptr, nullptr};
  ovrMirrorTexture mirrorTexture = nullptr;
  GLuint mirrorFBO = 0;
  long long frameIndex = 0;

  ovrSession session;
  ovrGraphicsLuid luid;
  ovrResult result = ovr_Create(&session, &luid);
  if (!OVR_SUCCESS(result))
    return retryCreate;

  if (Compare(luid, GetDefaultAdapterLuid())) // If luid that the Rift is on is not the default
                                              // adapter LUID...
  {
    VALIDATE(false, "OpenGL supports only the default graphics adapter.");
  }

  ovrHmdDesc hmdDesc = ovr_GetHmdDesc(session);

  // Setup Window and Graphics
  // Note: the mirror window can be any size, for this sample we use 1/2 the HMD resolution
  ovrSizei windowSize = {hmdDesc.Resolution.w / 2, hmdDesc.Resolution.h / 2};
  if (!Platform.InitDevice(windowSize.w, windowSize.h, reinterpret_cast<LUID*>(&luid)))
    goto Done;

  // Make eye render buffers
  for (int eye = 0; eye < 2; ++eye) {
    ovrSizei idealTextureSize =
        ovr_GetFovTextureSize(session, ovrEyeType(eye), hmdDesc.DefaultEyeFov[eye], 1);
    eyeRenderTexture[eye] = new OculusTextureBuffer(session, idealTextureSize, 1);

    if (!eyeRenderTexture[eye]->ColorTextureChain || !eyeRenderTexture[eye]->DepthTextureChain) {
      if (retryCreate)
        goto Done;
      VALIDATE(false, "Failed to create texture.");
    }
  }

  ovrMirrorTextureDesc desc;
  memset(&desc, 0, sizeof(desc));
  desc.Width = windowSize.w;
  desc.Height = windowSize.h;
  desc.Format = OVR_FORMAT_R8G8B8A8_UNORM_SRGB;

  // Create mirror texture and an FBO used to copy mirror texture to back buffer
  result = ovr_CreateMirrorTextureWithOptionsGL(session, &desc, &mirrorTexture);
  if (!OVR_SUCCESS(result)) {
    if (retryCreate)
      goto Done;
    VALIDATE(false, "Failed to create mirror texture.");
  }

  // Configure the mirror read buffer
  GLuint texId;
  ovr_GetMirrorTextureBufferGL(session, mirrorTexture, &texId);

  glGenFramebuffers(1, &mirrorFBO);
  glBindFramebuffer(GL_READ_FRAMEBUFFER, mirrorFBO);
  glFramebufferTexture2D(GL_READ_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texId, 0);
  glFramebufferRenderbuffer(GL_READ_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, 0);
  glBindFramebuffer(GL_READ_FRAMEBUFFER, 0);

  // Turn off vsync to let the compositor do its magic
  wglSwapIntervalEXT(0);

  // FloorLevel will give tracking poses where the floor height is 0
  ovr_SetTrackingOriginType(session, ovrTrackingOrigin_FloorLevel);

  {
    Application app{hmdDesc.Type == ovrHmdType::ovrHmd_CV1 ? Application::HMDType::Rift
                                                           : Application::HMDType::RiftS};

    double lastFrameTime = -1.0;

    struct ovrSceneMatrices {
      mat4 ViewMatrix[ovrGlProgram::MAX_VIEWS];
      mat4 ProjectionMatrix[ovrGlProgram::MAX_VIEWS];
    };

    ovrGlBuffer sceneMatrices(ovrGlBufferType_t::GLBUFFER_TYPE_UNIFORM, sizeof(ovrSceneMatrices));

    // Main loop
    while (Platform.HandleMessages()) {
      ovrSessionStatus sessionStatus;
      ovr_GetSessionStatus(session, &sessionStatus);
      if (sessionStatus.ShouldQuit) {
        // Because the application is requested to quit, should not request retry
        retryCreate = false;
        break;
      }
      if (sessionStatus.ShouldRecenter)
        ovr_RecenterTrackingOrigin(session);

      if (sessionStatus.IsVisible) {
        // Call ovr_GetRenderDesc each frame to get the ovrEyeRenderDesc, as the returned values
        // (e.g. HmdToEyePose) may change at runtime.
        ovrEyeRenderDesc eyeRenderDesc[2];
        eyeRenderDesc[0] = ovr_GetRenderDesc(session, ovrEye_Left, hmdDesc.DefaultEyeFov[0]);
        eyeRenderDesc[1] = ovr_GetRenderDesc(session, ovrEye_Right, hmdDesc.DefaultEyeFov[1]);

        // Get eye poses, feeding in correct IPD offset
        ovrPosef EyeRenderPose[2];
        ovrPosef HmdToEyePose[2] = {eyeRenderDesc[0].HmdToEyePose, eyeRenderDesc[1].HmdToEyePose};

        double sensorSampleTime; // sensorSampleTime is fed into the layer later
        ovr_GetEyePoses(
            session, frameIndex, ovrTrue, HmdToEyePose, EyeRenderPose, &sensorSampleTime);

        ovrTimewarpProjectionDesc posTimewarpProjectionDesc = {};

        double ftiming = ovr_GetPredictedDisplayTime(session, 0);
        if (lastFrameTime <= 0.0) {
          lastFrameTime = ftiming;
        }
        float deltaSeconds = ftiming - lastFrameTime;
        lastFrameTime = ftiming;

        ovrTrackingState trackingState = ovr_GetTrackingState(session, ftiming, ovrTrue);

        Application::FrameParams fParams;
        fParams.DeltaSeconds = deltaSeconds;
        fParams.EyePos = glm::make_vec3(&trackingState.HeadPose.ThePose.Position.x);
        fParams.EyeRot = glm::quat(
            trackingState.HeadPose.ThePose.Orientation.w,
            trackingState.HeadPose.ThePose.Orientation.x,
            trackingState.HeadPose.ThePose.Orientation.y,
            trackingState.HeadPose.ThePose.Orientation.z);

        bool controllerConnected[] = {false, false};
        ovrInputState InputStates[2];
        uint32_t connectedControllers = ovr_GetConnectedControllerTypes(session);
        if (connectedControllers && ovrControllerType_LTouch) {
          controllerConnected[0] = true;
          ovr_GetInputState(session, ovrControllerType_LTouch, &InputStates[0]);
        }
        if (connectedControllers && ovrControllerType_RTouch) {
          controllerConnected[1] = true;
          ovr_GetInputState(session, ovrControllerType_RTouch, &InputStates[1]);
        }

        for (uint32_t hand = 0; hand < 2; ++hand) {
          if (trackingState.HandStatusFlags[hand] & ovrStatus_OrientationTracked &&
              trackingState.HandStatusFlags[hand] & ovrStatus_PositionTracked &&
              controllerConnected[hand]) {
            fParams.Hands[hand].isActive = true;
            fParams.Hands[hand].Position =
                glm::make_vec3(&trackingState.HandPoses[hand].ThePose.Position.x);
            fParams.Hands[hand].Rotation =
                glm::make_quat(&trackingState.HandPoses[hand].ThePose.Orientation.x);

            fParams.Hands[hand].ButtonMask = 0;
            if (InputStates[hand].Buttons & ovrButton_X || InputStates[hand].Buttons & ovrButton_A)
              fParams.Hands[hand].ButtonMask |= ovrAvatarButton_One;
            if (InputStates[hand].Buttons & ovrButton_Y || InputStates[hand].Buttons & ovrButton_B)
              fParams.Hands[hand].ButtonMask |= ovrAvatarButton_Two;
            if (InputStates[hand].Buttons & ovrButton_Enter ||
                InputStates[hand].Buttons & ovrButton_Home)
              fParams.Hands[hand].ButtonMask |= ovrAvatarButton_Three;
            if (InputStates[hand].Buttons & ovrButton_LThumb ||
                InputStates[hand].Buttons & ovrButton_RThumb)
              fParams.Hands[hand].ButtonMask |= ovrAvatarButton_Joystick;

            fParams.Hands[hand].TouchMask = 0;
            if (InputStates[hand].Touches & ovrTouch_X || InputStates[hand].Touches & ovrTouch_A)
              fParams.Hands[hand].TouchMask |= ovrAvatarTouch_One;
            if (InputStates[hand].Touches & ovrTouch_Y || InputStates[hand].Touches & ovrTouch_B)
              fParams.Hands[hand].TouchMask |= ovrAvatarTouch_Two;
            if (InputStates[hand].Touches & ovrTouch_LThumb ||
                InputStates[hand].Touches & ovrTouch_RThumb)
              fParams.Hands[hand].TouchMask |= ovrAvatarTouch_Joystick;
            if (InputStates[hand].Touches & ovrTouch_LThumbRest ||
                InputStates[hand].Touches & ovrTouch_RThumbRest)
              fParams.Hands[hand].TouchMask |= ovrAvatarTouch_ThumbRest;
            if (InputStates[hand].Touches & ovrTouch_LIndexTrigger ||
                InputStates[hand].Touches & ovrTouch_RIndexTrigger)
              fParams.Hands[hand].TouchMask |= ovrAvatarTouch_Index;
            if (InputStates[hand].Touches & ovrTouch_LIndexPointing ||
                InputStates[hand].Touches & ovrTouch_RIndexPointing)
              fParams.Hands[hand].TouchMask |= ovrAvatarTouch_Pointing;
            if (InputStates[hand].Touches & ovrTouch_LThumbUp ||
                InputStates[hand].Touches & ovrTouch_RThumbUp)
              fParams.Hands[hand].TouchMask |= ovrAvatarTouch_ThumbUp;

            // For Avatar animations, always use the no deadzone versions. Since this app only uses
            // these for animation purposes, only bother sending the no deadzone versions. A real
            // app would need the normal versions for actual input as well.
            fParams.Hands[hand].JoystickX = InputStates[hand].ThumbstickNoDeadzone[hand].x;
            fParams.Hands[hand].JoystickY = InputStates[hand].ThumbstickNoDeadzone[hand].y;
            fParams.Hands[hand].HandTrigger = InputStates[hand].HandTriggerNoDeadzone[hand];
            fParams.Hands[hand].IndexTrigger = InputStates[hand].IndexTriggerNoDeadzone[hand];
          } else {
            fParams.Hands[hand].isActive = false;
          }
        }

        app.Frame(fParams);

        Matrix4f views[2];
        Matrix4f projs[2];

        // Update our scene matrices uniform Buffer
        ovrSceneMatrices* sceneMatricesMap = sceneMatrices.Map<ovrSceneMatrices>();
        for (int eye = 0; eye < 2; ++eye) {
          // Get view and projection matrices
          Matrix4f rollPitchYaw = Matrix4f(EyeRenderPose[eye].Orientation);
          Vector3f finalUp = rollPitchYaw.Transform(Vector3f(0, 1, 0));
          Vector3f finalForward = rollPitchYaw.Transform(Vector3f(0, 0, -1));
          Vector3f shiftedEyePos = EyeRenderPose[eye].Position;

          views[eye] = Matrix4f::LookAtRH(shiftedEyePos, shiftedEyePos + finalForward, finalUp);
          projs[eye] =
              ovrMatrix4f_Projection(hmdDesc.DefaultEyeFov[eye], 0.2f, 1000.0f, ovrProjection_None);

          sceneMatricesMap->ViewMatrix[eye] = transpose(glm::make_mat4((float*)&views[eye]));
          sceneMatricesMap->ProjectionMatrix[eye] = transpose(glm::make_mat4((float*)&projs[eye]));
        }
        sceneMatrices.Unmap();

        // Render Scene to Eye Buffers
        for (int eye = 0; eye < 2; ++eye) {
          // Switch to eye render target
          eyeRenderTexture[eye]->SetAndClearRenderSurface();

          posTimewarpProjectionDesc =
              ovrTimewarpProjectionDesc_FromProjection(projs[eye], ovrProjection_None);

          // Render world
          app.Render(eye, sceneMatrices);

          // Avoids an error when calling SetAndClearRenderSurface during next iteration.
          // Without this, during the next while loop iteration SetAndClearRenderSurface
          // would bind a framebuffer with an invalid COLOR_ATTACHMENT0 because the texture ID
          // associated with COLOR_ATTACHMENT0 had been unlocked by calling wglDXUnlockObjectsNV.
          eyeRenderTexture[eye]->UnsetRenderSurface();

          // Commit changes to the textures so they get picked up frame
          eyeRenderTexture[eye]->Commit();
        }

        // Do distortion rendering, Present and flush/sync

        ovrLayerEyeFovDepth ld = {};
        ld.Header.Type = ovrLayerType_EyeFovDepth;
        ld.Header.Flags = ovrLayerFlag_TextureOriginAtBottomLeft; // Because OpenGL.
        ld.ProjectionDesc = posTimewarpProjectionDesc;
        ld.SensorSampleTime = sensorSampleTime;

        for (int eye = 0; eye < 2; ++eye) {
          ld.ColorTexture[eye] = eyeRenderTexture[eye]->ColorTextureChain;
          ld.DepthTexture[eye] = eyeRenderTexture[eye]->DepthTextureChain;
          ld.Viewport[eye] = Recti(eyeRenderTexture[eye]->GetSize());
          ld.Fov[eye] = hmdDesc.DefaultEyeFov[eye];
          ld.RenderPose[eye] = EyeRenderPose[eye];
        }

        ovrLayerHeader* layers[] = {&ld.Header};
        result = ovr_SubmitFrame(session, frameIndex, nullptr, layers, (int)size(layers));
        // exit the rendering loop if submit returns an error, will retry on ovrError_DisplayLost
        if (!OVR_SUCCESS(result))
          goto Done;

        frameIndex++;
      }

      // Blit mirror texture to back buffer
      glBindFramebuffer(GL_READ_FRAMEBUFFER, mirrorFBO);
      glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
      GLint w = windowSize.w;
      GLint h = windowSize.h;
      glBlitFramebuffer(0, h, w, 0, 0, 0, w, h, GL_COLOR_BUFFER_BIT, GL_NEAREST);
      glBindFramebuffer(GL_READ_FRAMEBUFFER, 0);

      SwapBuffers(Platform.hDC);
    }
  }

Done:
  if (mirrorFBO)
    glDeleteFramebuffers(1, &mirrorFBO);
  if (mirrorTexture)
    ovr_DestroyMirrorTexture(session, mirrorTexture);
  for (int eye = 0; eye < 2; ++eye) {
    delete eyeRenderTexture[eye];
  }
  Platform.ReleaseDevice();
  ovr_Destroy(session);

  // Retry on ovrError_DisplayLost
  return retryCreate || (result == ovrError_DisplayLost);
}

//-------------------------------------------------------------------------------------
int WINAPI WinMain(_In_ HINSTANCE hinst, _In_opt_ HINSTANCE, _In_ LPSTR, _In_ int) {
  // Make sure our working directory is the exe directory. Not the case when running from
  // Visual Studio.
  {
    std::vector<wchar_t> exePath;
    const int maxPath = 32767; // NTFS max path length
    exePath.resize(maxPath);
    exePath.resize(GetModuleFileNameW(nullptr, exePath.data(), (DWORD)exePath.size()));

    fs::path path(exePath.data()); // Has exe name too
    fs::current_path(path.remove_filename());
  }

  // Initializes LibOVR, and the Rift
  ovrInitParams initParams = {
      ovrInit_RequestVersion | ovrInit_FocusAware, OVR_MINOR_VERSION, NULL, 0, 0};
  ovrResult result = ovr_Initialize(&initParams);
  VALIDATE(OVR_SUCCESS(result), "Failed to initialize libOVR.");

  VALIDATE(Platform.InitWindow(hinst, L"Oculus Room Tiny (GL)"), "Failed to open window.");

  Platform.Run(MainLoop);

  ovr_Shutdown();

  return (0);
}
